from typing import Self
from typing import TypeVar
from pyautd3.derive import builder
from pyautd3.derive import datagram
from pyautd3.derive import modulation
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.modulation import ModulationWithSamplingConfig
from pyautd3.driver.defined.freq import Freq
from pyautd3.driver.defined.freq import Hz
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.modulation.sampling_mode import ISamplingMode
from pyautd3.modulation.sampling_mode import SamplingModeExact
from pyautd3.modulation.sampling_mode import SamplingModeExactFloat
from pyautd3.modulation.sampling_mode import SamplingModeNearest
from pyautd3.native_methods.autd3capi_driver import ModulationPtr
from pyautd3.modulation.cache import Cache
from pyautd3.modulation.fir import Fir
from collections.abc import Iterable
from pyautd3.modulation.radiation_pressure import RadiationPressure
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold
from pyautd3.native_methods.autd3capi_driver import Segment, TransitionModeWrap
from pyautd3.driver.datagram.with_segment import DatagramWithSegment

T = TypeVar("T", int, float)

class Square(ModulationWithSamplingConfig):
    _mode: ISamplingMode
    def __private__init__(self: Square, mode: ISamplingMode) -> None: ...
    def __init__(self: Square, freq: Freq[T]) -> None: ...
    def _modulation_ptr(self: Square, ) -> ModulationPtr: ...
    def with_low(self: Square, low: int) -> Square: ...
    def with_high(self: Square, high: int) -> Square: ...
    def with_duty(self: Square, duty: float) -> Square: ...
    def with_cache(self: Square, ) -> Cache[Square]: ...
    def with_fir(self: Square, iterable: Iterable[float]) -> Fir[Square]: ...
    def with_radiation_pressure(self: Square, ) -> RadiationPressure[Square]: ...
    def with_timeout(self: Square, timeout: timedelta | None) -> DatagramWithTimeout[Square]: ...
    def with_parallel_threshold(self: Square, threshold: int | None) -> DatagramWithParallelThreshold[Square]: ...
    def with_segment(self: Square, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramWithSegment[Square]: ...
    @classmethod
    def nearest(cls, freq: Freq[float]) -> Square: ...
    @property
    def freq(self: Square) -> Freq[int] | Freq[float]: ...
    @property
    def low(self: Square) -> int: ...
    @property
    def high(self: Square) -> int: ...
    @property
    def duty(self: Square) -> float: ...



from typing import Self

from .emit_intensity import EmitIntensity
from .phase import Phase


class Drive:
    _phase: Phase
    _intensity: EmitIntensity

    def __init__(self: Self, value: "Drive | EmitIntensity | Phase | tuple") -> None:
        match value:
            case tuple():
                if len(value) != 2:  # noqa: PLR2004
                    raise TypeError
                match value[0], value[1]:
                    case EmitIntensity(), Phase():
                        self._intensity = value[0]
                        self._phase = value[1]
                    case Phase(), EmitIntensity():
                        self._intensity = value[1]
                        self._phase = value[0]
                    case _:
                        raise TypeError
            case Drive():
                self._phase = value.phase
                self._intensity = value.intensity
            case EmitIntensity():
                self._phase = Phase(0)
                self._intensity = value
            case Phase():
                self._phase = value
                self._intensity = EmitIntensity(0xFF)
            case _:
                raise TypeError

    @property
    def phase(self: Self) -> Phase:
        return self._phase

    @property
    def intensity(self: Self) -> EmitIntensity:
        return self._intensity

    @staticmethod
    def NULL() -> "Drive":  # noqa: N802
        return Drive((Phase(0), EmitIntensity.minimum()))



from typing import Self
import numpy as np
from numpy.typing import ArrayLike
import pyautd3.native_methods.autd3capi_driver as consts
from pyautd3.derive import builder



class AUTD3():
    TRANS_SPACING: float
    DEVICE_WIDTH: float
    DEVICE_HEIGHT: float
    NUM_TRANS_IN_X: int
    NUM_TRANS_IN_Y: int
    NUM_TRANS_IN_UNIT: int
    def __init__(self: AUTD3, pos: ArrayLike) -> None: ...
    def with_rotation(self: AUTD3, rotation: np.ndarray) -> AUTD3: ...
    @property
    def position(self: AUTD3) -> np.ndarray: ...
    @property
    def rotation(self: AUTD3) -> np.ndarray: ...



from collections.abc import Callable
from ctypes import CFUNCTYPE
from ctypes import c_uint8
from ctypes import c_uint16
from ctypes import c_void_p
from threading import Lock
from typing import Self
from pyautd3.derive import datagram
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.firmware.fpga.phase import Phase
from pyautd3.driver.geometry import Geometry
from pyautd3.driver.geometry.device import Device
from pyautd3.driver.geometry.transducer import Transducer
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr
from pyautd3.native_methods.autd3capi_driver import GeometryPtr
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold



class PhaseCorrection(Datagram):
    _cache: dict[int, Callable[[Transducer], Phase]]
    _lock: Lock
    def __init__(self: PhaseCorrection, f: Callable[[Device], Callable[[Transducer], Phase]]) -> None: ...
    def _datagram_ptr(self: PhaseCorrection, geometry: Geometry) -> DatagramPtr: ...
    def with_timeout(self: PhaseCorrection, timeout: timedelta | None) -> DatagramWithTimeout[PhaseCorrection]: ...
    def with_parallel_threshold(self: PhaseCorrection, threshold: int | None) -> DatagramWithParallelThreshold[PhaseCorrection]: ...



from .derive_builder import builder
from .derive_gain import gain
from .derive_datagram import datagram, datagram_with_segment
from .derive_modulation import modulation

__all__ = ["builder", "gain", "datagram", "modulation", "datagram_with_segment"]



from typing import Self

from pyautd3.driver.link import Link, LinkBuilder
from pyautd3.native_methods.autd3capi import ControllerPtr
from pyautd3.native_methods.autd3capi import (
    NativeMethods as LinkNop,
)
from pyautd3.native_methods.autd3capi_driver import HandlePtr, LinkBuilderPtr, LinkPtr


class Nop(Link):
    class _Builder(LinkBuilder["Nop"]):
        def __init__(self: Self) -> None:
            pass

        def _link_builder_ptr(self: Self) -> LinkBuilderPtr:
            return LinkNop().link_nop()

        def _resolve_link(self: Self, handle: HandlePtr, ptr: ControllerPtr) -> "Nop":
            return Nop(handle, LinkNop().link_get(ptr))

    def __init__(self: Self, handle: HandlePtr, ptr: LinkPtr) -> None:
        super().__init__(handle, ptr)

    @staticmethod
    def builder() -> _Builder:
        return Nop._Builder()



import platform
from typing import Self
from pyautd3.derive import builder
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import SpinStrategyTag as SpinStrategy
from pyautd3.native_methods.autd3capi_driver import TimerStrategyWrap
from pyautd3.native_methods.utils import ConstantADT



class StdSleeper():
    def __init__(self: StdSleeper, ) -> None: ...
    @property
    def timer_resolution(self: StdSleeper) -> int | None: ...

class SpinSleeper():
    def __init__(self: SpinSleeper, native_accuracy_ns: int | None = None) -> None: ...
    def with_spin_strategy(self: SpinSleeper, spin_strategy: SpinStrategy) -> SpinSleeper: ...
    @property
    def native_accuracy_ns(self: SpinSleeper) -> int: ...
    @property
    def spin_strategy(self: SpinSleeper) -> SpinStrategy: ...

class AsyncSleeper():
    def __init__(self: AsyncSleeper, ) -> None: ...
    @property
    def timer_resolution(self: AsyncSleeper) -> int | None: ...

class WaitableSleeper():
    def __init__(self: WaitableSleeper, ) -> None: ...

class TimerStrategy():
    def __new__(cls, ) -> TimerStrategy: ...
    @staticmethod
    def Std(sleeper: StdSleeper) -> TimerStrategyWrap: ...
    @staticmethod
    def Spin(sleeper: SpinSleeper) -> TimerStrategyWrap: ...
    @staticmethod
    def Async(sleeper: AsyncSleeper) -> TimerStrategyWrap: ...
    @staticmethod
    def Waitable(_sleeper: WaitableSleeper) -> TimerStrategyWrap: ...



# This file is autogenerated
import threading
import ctypes
import os
from pyautd3.native_methods.structs import Vector3, Quaternion, FfiFuture, LocalFfiFuture
from pyautd3.native_methods.autd3_driver import SamplingConfig, LoopBehavior, SyncMode, GainSTMMode, GPIOOut, GPIOIn, Segment, SilencerTarget, Drive
from enum import IntEnum


class DebugTypeTag(IntEnum):
    None_ = 0
    BaseSignal = 1
    Thermo = 2
    ForceFan = 3
    Sync = 4
    ModSegment = 5
    ModIdx = 6
    StmSegment = 7
    StmIdx = 8
    IsStmMode = 9
    PwmOut = 10
    Direct = 11
    SysTimeEq = 12

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class TransitionModeTag(IntEnum):
    SyncIdx = 0
    SysTime = 1
    Gpio = 2
    Ext = 3
    Immediate = 4
    None_ = 0xFF

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class DynWindow(IntEnum):
    Rectangular = 0
    Blackman = 1

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class AUTDStatus(IntEnum):
    AUTDTrue = 0
    AUTDFalse = 1
    AUTDErr = 2

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class TimerStrategyTag(IntEnum):
    Std = 0
    Spin = 1
    Async = 2
    Waitable = 3

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class SpinStrategyTag(IntEnum):
    YieldThread = 0
    SpinLoopHint = 1

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class ConstPtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class ControllerPtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class ControllerBuilderPtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class DatagramPtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class GainPtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class GeometryPtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class DevicePtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class TransducerPtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class LinkBuilderPtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class LinkPtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class SyncLinkBuilderPtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class ModulationPtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class RuntimePtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class HandlePtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class FociSTMPtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class GainSTMPtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class DebugTypeWrap(ctypes.Structure):
    _fields_ = [("ty", ctypes.c_uint8), ("value", ctypes.c_uint64)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, DebugTypeWrap) and self._fields_ == other._fields_  # pragma: no cover


class TransitionModeWrap(ctypes.Structure):
    _fields_ = [("tag", ctypes.c_uint8), ("value", ctypes.c_uint64)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, TransitionModeWrap) and self._fields_ == other._fields_  # pragma: no cover


class ResultDatagram(ctypes.Structure):
    _fields_ = [("result", DatagramPtr), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultDatagram) and self._fields_ == other._fields_  # pragma: no cover


class ResultLinkBuilder(ctypes.Structure):
    _fields_ = [("result", LinkBuilderPtr), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultLinkBuilder) and self._fields_ == other._fields_  # pragma: no cover


class ResultSyncLinkBuilder(ctypes.Structure):
    _fields_ = [("result", LinkBuilderPtr), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultSyncLinkBuilder) and self._fields_ == other._fields_  # pragma: no cover


class ResultModulation(ctypes.Structure):
    _fields_ = [("result", ModulationPtr), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultModulation) and self._fields_ == other._fields_  # pragma: no cover


class ResultFociSTM(ctypes.Structure):
    _fields_ = [("result", FociSTMPtr), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultFociSTM) and self._fields_ == other._fields_  # pragma: no cover


class ResultGainSTM(ctypes.Structure):
    _fields_ = [("result", GainSTMPtr), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultGainSTM) and self._fields_ == other._fields_  # pragma: no cover


class DynSincInterpolator(ctypes.Structure):
    _fields_ = [("window", ctypes.c_uint32), ("window_size", ctypes.c_uint32)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, DynSincInterpolator) and self._fields_ == other._fields_  # pragma: no cover


class ResultStatus(ctypes.Structure):
    _fields_ = [("result", ctypes.c_uint8), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultStatus) and self._fields_ == other._fields_  # pragma: no cover


class ResultSamplingConfig(ctypes.Structure):
    _fields_ = [("result", SamplingConfig), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultSamplingConfig) and self._fields_ == other._fields_  # pragma: no cover


class TimerStrategyWrap(ctypes.Structure):
    _fields_ = [("tag", ctypes.c_uint8), ("value", ctypes.c_uint32), ("spin_strategy", ctypes.c_uint8)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, TimerStrategyWrap) and self._fields_ == other._fields_  # pragma: no cover


NUM_TRANS_IN_UNIT: int = 249

NUM_TRANS_IN_X: int = 18

NUM_TRANS_IN_Y: int = 14

TRANS_SPACING_MM: float = 10.16

DEVICE_HEIGHT_MM: float = 151.4

DEVICE_WIDTH_MM: float = 192.0




from typing import Self

from pyautd3.driver.defined import Angle
from pyautd3.native_methods.autd3capi import NativeMethods as Base


class Phase:
    _value: int

    def __init__(self: Self, phase: "int | Angle | Phase") -> None:
        match phase:
            case int():
                self._value = phase
            case Angle():
                self._value = int(Base().phase_from_rad(phase.radian))
            case Phase():
                self._value = phase._value
            case _:
                raise TypeError

    @property
    def value(self: Self) -> int:
        return int(self._value)

    @property
    def radian(self: Self) -> float:
        return float(Base().phase_to_rad(self.value))

    def __eq__(self: Self, other: object) -> bool:
        return isinstance(other, Phase) and self.value == other.value

    def __str__(self: Self) -> str:
        return f"Phase({self.value})"

    def __repr__(self: Self) -> str:
        return f"Phase({self.value})"



from collections.abc import Callable
from ctypes import CFUNCTYPE, c_uint8, c_uint16, c_void_p
from threading import Lock
from typing import Self

from pyautd3.derive import datagram
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.firmware.fpga.phase import Phase
from pyautd3.driver.geometry import Geometry
from pyautd3.driver.geometry.device import Device
from pyautd3.driver.geometry.transducer import Transducer
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, GeometryPtr


@datagram
class PhaseCorrection(Datagram):
    _cache: dict[int, Callable[[Transducer], Phase]]
    _lock: Lock

    def __init__(self: Self, f: Callable[[Device], Callable[[Transducer], Phase]]) -> None:
        super().__init__()
        self._cache = {}
        self._lock = Lock()

        def f_native(_context: c_void_p, geometry_ptr: GeometryPtr, dev_idx: int, tr_idx: int) -> int:
            if dev_idx not in self._cache:
                with self._lock:
                    self._cache[dev_idx] = f(Device(dev_idx, geometry_ptr))
            return self._cache[dev_idx](Transducer(tr_idx, dev_idx, Base().device(geometry_ptr, dev_idx))).value

        self._f_native = CFUNCTYPE(c_uint8, c_void_p, GeometryPtr, c_uint16, c_uint8)(f_native)

    def _datagram_ptr(self: Self, geometry: Geometry) -> DatagramPtr:
        return Base().datagram_phase_corr(self._f_native, None, geometry._geometry_ptr)  # type: ignore[arg-type]



# This file is autogenerated
import threading
import ctypes
import os
from pyautd3.native_methods.structs import Vector3, Quaternion, FfiFuture, LocalFfiFuture
from pyautd3.native_methods.autd3_driver import SamplingConfig, LoopBehavior, SyncMode, GainSTMMode, GPIOOut, GPIOIn, Segment, SilencerTarget, Drive
from pyautd3.native_methods.autd3capi_driver import DynSincInterpolator, ResultModulation, ResultStatus



class Singleton(type):
    _instances = {}  # type: ignore
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances: # pragma: no cover
                    cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class NativeMethods(metaclass=Singleton):

    def init_dll(self, bin_location: str, bin_prefix: str, bin_ext: str):
        self.dll = ctypes.CDLL(os.path.join(bin_location, f'{bin_prefix}autd3capi_modulation_audio_file{bin_ext}'))

        self.dll.AUTDModulationAudioFileTracingInit.argtypes = [] 
        self.dll.AUTDModulationAudioFileTracingInit.restype = None

        self.dll.AUTDModulationAudioFileTracingInitWithFile.argtypes = [ctypes.c_char_p] 
        self.dll.AUTDModulationAudioFileTracingInitWithFile.restype = ResultStatus

        self.dll.AUTDModulationAudioFileWav.argtypes = [ctypes.c_char_p, LoopBehavior]  # type: ignore 
        self.dll.AUTDModulationAudioFileWav.restype = ResultModulation

        self.dll.AUTDModulationAudioFileWavWithResample.argtypes = [ctypes.c_char_p, LoopBehavior, SamplingConfig, DynSincInterpolator]  # type: ignore 
        self.dll.AUTDModulationAudioFileWavWithResample.restype = ResultModulation

        self.dll.AUTDModulationAudioFileRawPCM.argtypes = [ctypes.c_char_p, SamplingConfig, LoopBehavior]  # type: ignore 
        self.dll.AUTDModulationAudioFileRawPCM.restype = ResultModulation

        self.dll.AUTDModulationAudioFileRawPCMWithResample.argtypes = [ctypes.c_char_p, LoopBehavior, ctypes.c_float, SamplingConfig, DynSincInterpolator]  # type: ignore 
        self.dll.AUTDModulationAudioFileRawPCMWithResample.restype = ResultModulation

        self.dll.AUTDModulationAudioFileCsv.argtypes = [ctypes.c_char_p, SamplingConfig, ctypes.c_uint8, LoopBehavior]  # type: ignore 
        self.dll.AUTDModulationAudioFileCsv.restype = ResultModulation

        self.dll.AUTDModulationAudioFileCsvWithResample.argtypes = [ctypes.c_char_p, ctypes.c_uint8, LoopBehavior, ctypes.c_float, SamplingConfig, DynSincInterpolator]  # type: ignore 
        self.dll.AUTDModulationAudioFileCsvWithResample.restype = ResultModulation

    def modulation_audio_file_tracing_init(self) -> None:
        return self.dll.AUTDModulationAudioFileTracingInit()

    def modulation_audio_file_tracing_init_with_file(self, path: bytes) -> ResultStatus:
        return self.dll.AUTDModulationAudioFileTracingInitWithFile(path)

    def modulation_audio_file_wav(self, path: bytes, loop_behavior: LoopBehavior) -> ResultModulation:
        return self.dll.AUTDModulationAudioFileWav(path, loop_behavior)

    def modulation_audio_file_wav_with_resample(self, path: bytes, loop_behavior: LoopBehavior, target: SamplingConfig, resample: DynSincInterpolator) -> ResultModulation:
        return self.dll.AUTDModulationAudioFileWavWithResample(path, loop_behavior, target, resample)

    def modulation_audio_file_raw_pcm(self, path: bytes, config: SamplingConfig, loop_behavior: LoopBehavior) -> ResultModulation:
        return self.dll.AUTDModulationAudioFileRawPCM(path, config, loop_behavior)

    def modulation_audio_file_raw_pcm_with_resample(self, path: bytes, loop_behavior: LoopBehavior, src: float, target: SamplingConfig, resample: DynSincInterpolator) -> ResultModulation:
        return self.dll.AUTDModulationAudioFileRawPCMWithResample(path, loop_behavior, src, target, resample)

    def modulation_audio_file_csv(self, path: bytes, config: SamplingConfig, deliminator: int, loop_behavior: LoopBehavior) -> ResultModulation:
        return self.dll.AUTDModulationAudioFileCsv(path, config, deliminator, loop_behavior)

    def modulation_audio_file_csv_with_resample(self, path: bytes, deliminator: int, loop_behavior: LoopBehavior, src: float, target: SamplingConfig, resample: DynSincInterpolator) -> ResultModulation:
        return self.dll.AUTDModulationAudioFileCsvWithResample(path, deliminator, loop_behavior, src, target, resample)



from typing import Self

from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr

__all__ = []  # type: ignore[var-annotated]


class DatagramTuple(Datagram):
    _d1: Datagram
    _d2: Datagram

    def __init__(self: Self, d: tuple[Datagram, Datagram]) -> None:
        self._d1, self._d2 = d

    def _datagram_ptr(self: Self, g: Geometry) -> DatagramPtr:
        return Base().datagram_tuple(self._d1._datagram_ptr(g), self._d2._datagram_ptr(g))



from datetime import timedelta

from forbiddenfruit import curse  # type: ignore[import-untyped]

from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.datagram.datagram_tuple import DatagramTuple
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout

from .clear import Clear
from .debug import DebugSettings, DebugType
from .force_fan import ForceFan
from .phase_corr import PhaseCorrection
from .pulse_width_encoder import PulseWidthEncoder
from .reads_fpga_state import ReadsFPGAState
from .segment import SwapSegment
from .silencer import FixedCompletionTime, FixedUpdateRate, Silencer
from .stm import FociSTM, GainSTM, GainSTMMode
from .synchronize import Synchronize

__all__ = [
    "Clear",
    "Silencer",
    "FixedCompletionTime",
    "FixedUpdateRate",
    "DebugSettings",
    "DebugType",
    "ReadsFPGAState",
    "ForceFan",
    "Datagram",
    "Synchronize",
    "GainSTM",
    "FociSTM",
    "GainSTMMode",
    "SwapSegment",
    "PulseWidthEncoder",
    "PhaseCorrection",
]


def __with_parallel_threshold(self: tuple[Datagram, Datagram], threshold: int) -> DatagramWithParallelThreshold[DatagramTuple]:
    return DatagramWithParallelThreshold(DatagramTuple(self), threshold)


curse(tuple, "with_parallel_threshold", __with_parallel_threshold)


def __with_timeout(self: tuple[Datagram, Datagram], timeout: timedelta) -> DatagramWithTimeout[DatagramTuple]:
    return DatagramWithTimeout(DatagramTuple(self), timeout)


curse(tuple, "with_timeout", __with_timeout)



# This file is autogenerated
import threading
import ctypes
import os
from pyautd3.native_methods.structs import Vector3, Quaternion, FfiFuture, LocalFfiFuture
from pyautd3.native_methods.autd3_driver import SamplingConfig, LoopBehavior, SyncMode, GainSTMMode, GPIOOut, GPIOIn, Segment, SilencerTarget, Drive
from pyautd3.native_methods.autd3capi_driver import LinkBuilderPtr, ResultLinkBuilder, ResultStatus



class Singleton(type):
    _instances = {}  # type: ignore
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances: # pragma: no cover
                    cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class NativeMethods(metaclass=Singleton):

    def init_dll(self, bin_location: str, bin_prefix: str, bin_ext: str):
        self.dll = ctypes.CDLL(os.path.join(bin_location, f'{bin_prefix}autd3capi_link_twincat{bin_ext}'))

        self.dll.AUTDLinkTwinCATTracingInit.argtypes = [] 
        self.dll.AUTDLinkTwinCATTracingInit.restype = None

        self.dll.AUTDLinkTwinCATTracingInitWithFile.argtypes = [ctypes.c_char_p] 
        self.dll.AUTDLinkTwinCATTracingInitWithFile.restype = ResultStatus

        self.dll.AUTDLinkTwinCAT.argtypes = [] 
        self.dll.AUTDLinkTwinCAT.restype = LinkBuilderPtr

        self.dll.AUTDLinkRemoteTwinCAT.argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p] 
        self.dll.AUTDLinkRemoteTwinCAT.restype = ResultLinkBuilder

    def link_twin_cat_tracing_init(self) -> None:
        return self.dll.AUTDLinkTwinCATTracingInit()

    def link_twin_cat_tracing_init_with_file(self, path: bytes) -> ResultStatus:
        return self.dll.AUTDLinkTwinCATTracingInitWithFile(path)

    def link_twin_cat(self) -> LinkBuilderPtr:
        return self.dll.AUTDLinkTwinCAT()

    def link_remote_twin_cat(self, server_ams_net_id: bytes, server_ip: bytes, client_ams_net_id: bytes) -> ResultLinkBuilder:
        return self.dll.AUTDLinkRemoteTwinCAT(server_ams_net_id, server_ip, client_ams_net_id)



import ctypes
from typing import Self
import numpy as np


class FfiFuture(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_byte * 24)]


class LocalFfiFuture(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_byte * 24)]


class Vector3(ctypes.Structure):
    _fields_ = [("x", ctypes.c_float), ("y", ctypes.c_float), ("z", ctypes.c_float)]

    def __init__(self: Self, v: np.ndarray) -> None:
        self.x = v[0]
        self.y = v[1]
        self.z = v[2]

    def ndarray(self):
        return np.array([self.x, self.y, self.z])


class Quaternion(ctypes.Structure):
    _fields_ = [("x", ctypes.c_float), ("y", ctypes.c_float), ("z", ctypes.c_float), ("w", ctypes.c_float)]

    def __init__(self: Self, v: np.ndarray) -> None:
        self.w = v[0]
        self.x = v[1]
        self.y = v[2]
        self.z = v[3]

    def ndarray(self):
        return np.array([self.w, self.x, self.y, self.z])



from datetime import timedelta
from typing import Generic
from typing import Self
from typing import TypeVar
from pyautd3.derive import builder
from pyautd3.derive import datagram
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.driver.datagram.silencer.fixed_completion_time import FixedCompletionTime
from pyautd3.driver.datagram.silencer.fixed_update_rate import FixedUpdateRate
from pyautd3.driver.datagram.stm.foci import FociSTM
from pyautd3.driver.datagram.stm.gain import GainSTM
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi_driver import DatagramPtr
from pyautd3.native_methods.autd3capi_driver import SilencerTarget
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold

T = TypeVar("T", FixedCompletionTime, FixedUpdateRate)

class Silencer(Datagram, Generic[T]):
    _inner: T
    _strict_mode: bool
    def __init__(self: Silencer[T], config: T | None = None) -> None: ...
    def with_strict_mode(self: Silencer[T], mode: bool) -> Silencer[T]: ...
    def is_valid(self: Silencer[T], target: Modulation | FociSTM | GainSTM) -> bool: ...
    def _datagram_ptr(self: Silencer[T], _: Geometry) -> DatagramPtr: ...
    def with_target(self: Silencer[T], target: SilencerTarget) -> Silencer[T]: ...
    def with_timeout(self: Silencer[T], timeout: timedelta | None) -> DatagramWithTimeout[Silencer[T]]: ...
    def with_parallel_threshold(self: Silencer[T], threshold: int | None) -> DatagramWithParallelThreshold[Silencer[T]]: ...
    @staticmethod
    def disable() -> Silencer[FixedCompletionTime]: ...
    @property
    def target(self: Silencer[T]) -> SilencerTarget: ...



import ctypes
from collections.abc import Iterable
from typing import Self
import numpy as np
from pyautd3.derive import builder
from pyautd3.derive import datagram
from pyautd3.derive import gain
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.geometry import Geometry
from pyautd3.gain.holo.amplitude import Amplitude
from pyautd3.gain.holo.backend import Backend
from pyautd3.gain.holo.constraint import EmissionConstraint
from pyautd3.gain.holo.holo import HoloWithBackend
from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.native_methods.structs import Vector3
from pyautd3.gain.cache import Cache
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold
from pyautd3.native_methods.autd3capi_driver import Segment, TransitionModeWrap
from pyautd3.driver.datagram.with_segment import DatagramWithSegment



class GS(HoloWithBackend[GS]):
    def __init__(self: GS, backend: Backend, iterable: Iterable[tuple[np.ndarray, Amplitude]]) -> None: ...
    def _gain_ptr(self: GS, _: Geometry) -> GainPtr: ...
    def with_repeat(self: GS, repeat: int) -> GS: ...
    def with_cache(self: GS, ) -> Cache[GS]: ...
    def with_timeout(self: GS, timeout: timedelta | None) -> DatagramWithTimeout[GS]: ...
    def with_parallel_threshold(self: GS, threshold: int | None) -> DatagramWithParallelThreshold[GS]: ...
    def with_segment(self: GS, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramWithSegment[GS]: ...
    @property
    def repeat(self: GS) -> int: ...



import asyncio
import ctypes
from collections.abc import Callable
from datetime import timedelta
from types import TracebackType
from typing import Self

import numpy as np
import polars as pl

from pyautd3.controller.controller import Controller, _Builder
from pyautd3.driver.geometry.geometry import Geometry
from pyautd3.driver.link import Link
from pyautd3.native_methods.autd3capi import ControllerPtr, RuntimePtr
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import GeometryPtr, HandlePtr, LinkPtr
from pyautd3.native_methods.autd3capi_emulator import EmulatorPtr, RecordPtr, SoundFieldPtr
from pyautd3.native_methods.autd3capi_emulator import NativeMethods as Emu
from pyautd3.native_methods.autd3capi_emulator import Range as Range_
from pyautd3.native_methods.autd3capi_emulator import RecordOption as RecordOption_
from pyautd3.native_methods.utils import _validate_ptr, _validate_status


class Range:
    _inner: Range_

    def __init__(
        self: Self,
        *,
        x_start: float,
        x_end: float,
        y_start: float,
        y_end: float,
        z_start: float,
        z_end: float,
        resolution: float,
    ) -> None:
        self._inner = Range_(
            x_start,
            x_end,
            y_start,
            y_end,
            z_start,
            z_end,
            resolution,
        )


class RecordOption:
    _inner: RecordOption_

    def __init__(
        self: Self,
        *,
        sound_speed: float | None = None,
        time_step_ns: int | None = None,
        print_progress: bool | None = None,
        memory_limits_hint_mb: int | None = None,
        gpu: bool | None = None,
    ) -> None:
        self._inner = RecordOption_(
            sound_speed if sound_speed is not None else 340e3,
            time_step_ns if time_step_ns is not None else 1000,
            print_progress if print_progress is not None else False,
            memory_limits_hint_mb if memory_limits_hint_mb is not None else 128,
            gpu if gpu is not None else False,
        )


class SoundField:
    _ptr: SoundFieldPtr
    _handle: HandlePtr

    def __init__(self: Self, ptr: SoundFieldPtr, handle: HandlePtr) -> None:
        self._ptr = ptr
        self._handle = handle

    def __del__(self: Self) -> None:
        self._dispose()

    def _dispose(self: Self) -> None:
        if self._ptr._0 is not None:  # pragma: no cover
            Emu().emulator_sound_field_free(self._ptr)
            self._ptr._0 = None

    def skip(self: Self, duration: timedelta) -> Self:
        _validate_status(
            Base().wait_local_result_status(
                self._handle,
                Emu().emulator_sound_field_skip(self._ptr, int(duration.total_seconds() * 1000 * 1000 * 1000)),
            ),
        )
        return self

    def next(self: Self, duration: timedelta) -> pl.DataFrame:
        n = int(Emu().emulator_sound_field_time_len(self._ptr, int(duration.total_seconds() * 1000 * 1000 * 1000)))
        points_len = int(Emu().emulator_sound_field_points_len(self._ptr))
        time = np.zeros(n, dtype=np.uint64)

        x = np.zeros(points_len, dtype=np.float32)
        y = np.zeros(points_len, dtype=np.float32)
        z = np.zeros(points_len, dtype=np.float32)
        Emu().emulator_sound_field_get_x(self._ptr, x.ctypes.data_as(ctypes.POINTER(ctypes.c_float)))  # type: ignore[arg-type]
        Emu().emulator_sound_field_get_y(self._ptr, y.ctypes.data_as(ctypes.POINTER(ctypes.c_float)))  # type: ignore[arg-type]
        Emu().emulator_sound_field_get_z(self._ptr, z.ctypes.data_as(ctypes.POINTER(ctypes.c_float)))  # type: ignore[arg-type]

        v = np.zeros([n, points_len], dtype=np.float32)
        _validate_status(
            Base().wait_local_result_status(
                self._handle,
                Emu().emulator_sound_field_next(
                    self._ptr,
                    int(duration.total_seconds() * 1000 * 1000 * 1000),
                    time.ctypes.data_as(ctypes.POINTER(ctypes.c_ulonglong)),  # type: ignore[arg-type]
                    ctypes.cast(
                        (ctypes.POINTER(ctypes.c_float) * n)(
                            *(ctypes.cast(r, ctypes.POINTER(ctypes.c_float)) for r in np.ctypeslib.as_ctypes(v)),  # type: ignore[arg-type]
                        ),
                        ctypes.POINTER(ctypes.POINTER(ctypes.c_float)),
                    ),  # type: ignore[arg-type]
                ),
            ),
        )
        return pl.DataFrame(
            {
                "x[mm]": x,
                "y[mm]": y,
                "z[mm]": z,
                **{s.name: s for s in (pl.Series(name=f"p[Pa]@{time[i]}[ns]", values=v[i]) for i in range(n))},
            },
        )

    async def next_async(self: Self, duration: timedelta) -> pl.DataFrame:
        future: asyncio.Future = asyncio.Future()
        loop = asyncio.get_event_loop()

        n = int(Emu().emulator_sound_field_time_len(self._ptr, int(duration.total_seconds() * 1000 * 1000 * 1000)))
        points_len = int(Emu().emulator_sound_field_points_len(self._ptr))
        time = np.zeros(n, dtype=np.uint64)

        x = np.zeros(points_len, dtype=np.float32)
        y = np.zeros(points_len, dtype=np.float32)
        z = np.zeros(points_len, dtype=np.float32)
        Emu().emulator_sound_field_get_x(self._ptr, x.ctypes.data_as(ctypes.POINTER(ctypes.c_float)))  # type: ignore[arg-type]
        Emu().emulator_sound_field_get_y(self._ptr, y.ctypes.data_as(ctypes.POINTER(ctypes.c_float)))  # type: ignore[arg-type]
        Emu().emulator_sound_field_get_z(self._ptr, z.ctypes.data_as(ctypes.POINTER(ctypes.c_float)))  # type: ignore[arg-type]

        v = np.zeros([n, points_len], dtype=np.float32)
        ffi_future = Emu().emulator_sound_field_next(
            self._ptr,
            int(duration.total_seconds() * 1000 * 1000 * 1000),
            time.ctypes.data_as(ctypes.POINTER(ctypes.c_ulonglong)),  # type: ignore[arg-type]
            ctypes.cast(
                (ctypes.POINTER(ctypes.c_float) * n)(
                    *(ctypes.cast(r, ctypes.POINTER(ctypes.c_float)) for r in np.ctypeslib.as_ctypes(v)),  # type: ignore[arg-type]
                ),
                ctypes.POINTER(ctypes.POINTER(ctypes.c_float)),
            ),  # type: ignore[arg-type]
        )
        loop.call_soon(
            lambda *_: future.set_result(
                Base().wait_local_result_status(
                    self._handle,
                    ffi_future,
                ),
            ),
        )
        _validate_status(await future)
        return pl.DataFrame(
            {
                "x[mm]": x,
                "y[mm]": y,
                "z[mm]": z,
                **{s.name: s for s in (pl.Series(name=f"p[Pa]@{time[i]}[ns]", values=v[i]) for i in range(n))},
            },
        )


class Recorder(Link):
    def __init__(self: Self, handle: HandlePtr, ptr: LinkPtr) -> None:
        super().__init__(handle, ptr)

    def tick(self: Self, tick: timedelta) -> None:
        _validate_status(
            Emu().emulator_tick_ns(self._ptr, int(tick.total_seconds() * 1000 * 1000 * 1000)),
        )


Controller[Recorder].tick = lambda self, tick: self._link.tick(tick)  # type: ignore[attr-defined,method-assign]


class Record:
    _ptr: RecordPtr
    _handle: HandlePtr

    def __init__(self: Self, ptr: RecordPtr, handle: HandlePtr) -> None:
        self._ptr = ptr
        self._handle = handle

    def __del__(self: Self) -> None:
        self._dispose()

    def _dispose(self: Self) -> None:
        if self._ptr._0 is not None:  # pragma: no cover
            Emu().emulator_record_free(self._ptr)
            self._ptr._0 = None

    def drive(self: Self) -> pl.DataFrame:
        n = int(Emu().emulator_record_drive_len(self._ptr))
        time = np.zeros(n, dtype=np.uint64)
        Emu().emulator_record_drive_time(
            self._ptr,
            time.ctypes.data_as(ctypes.POINTER(ctypes.c_ulonglong)),  # type: ignore[arg-type]
        )
        dev_num = int(Emu().emulator_record_num_devices(self._ptr))
        pulsewidth = np.zeros(n, dtype=np.uint8)
        phase = np.zeros(n, dtype=np.uint8)

        num_cols = sum(2 * int(Emu().emulator_record_num_transducers(self._ptr, dev_idx)) for dev_idx in range(dev_num))
        series = np.zeros([num_cols], dtype=pl.Series)
        i = 0
        for dev_idx in range(dev_num):
            for tr_idx in range(int(Emu().emulator_record_num_transducers(self._ptr, dev_idx))):
                Emu().emulator_record_drive_phase(
                    self._ptr,
                    dev_idx,
                    tr_idx,
                    phase.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8)),  # type: ignore[arg-type]
                )
                Emu().emulator_record_drive_pulse_width(
                    self._ptr,
                    dev_idx,
                    tr_idx,
                    pulsewidth.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8)),  # type: ignore[arg-type]
                )
                series[i] = pl.Series(name=f"phase_{dev_idx}_{tr_idx}", values=phase)
                series[i + 1] = pl.Series(name=f"pulsewidth_{dev_idx}_{tr_idx}", values=pulsewidth)
                i += 2

        return pl.DataFrame(
            {
                "time[ns]": time,
                **{s.name: s for s in series},
            },
        )

    def output_voltage(self: Self) -> pl.DataFrame:
        n = int(Emu().emulator_record_output_len(self._ptr))
        time = np.zeros(n, dtype=np.uint64)
        Emu().emulator_record_output_time(
            self._ptr,
            time.ctypes.data_as(ctypes.POINTER(ctypes.c_ulonglong)),  # type: ignore[arg-type]
        )
        dev_num = int(Emu().emulator_record_num_devices(self._ptr))
        v = np.zeros(n, dtype=np.float32)

        num_cols = sum(int(Emu().emulator_record_num_transducers(self._ptr, dev_idx)) for dev_idx in range(dev_num))
        series = np.zeros([num_cols], dtype=pl.Series)
        i = 0
        for dev_idx in range(dev_num):
            for tr_idx in range(int(Emu().emulator_record_num_transducers(self._ptr, dev_idx))):
                Emu().emulator_record_output_voltage(
                    self._ptr,
                    dev_idx,
                    tr_idx,
                    v.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),  # type: ignore[arg-type]
                )
                series[i] = pl.Series(name=f"voltage_{dev_idx}_{tr_idx}[V]", values=v)
                i += 1

        return pl.DataFrame(
            {
                "time[25us/256]": time,
                **{s.name: s for s in series},
            },
        )

    def output_ultrasound(self: Self) -> pl.DataFrame:
        n = int(Emu().emulator_record_output_len(self._ptr))
        time = np.zeros(n, dtype=np.uint64)
        Emu().emulator_record_output_time(
            self._ptr,
            time.ctypes.data_as(ctypes.POINTER(ctypes.c_ulonglong)),  # type: ignore[arg-type]
        )
        dev_num = int(Emu().emulator_record_num_devices(self._ptr))
        v = np.zeros(n, dtype=np.float32)

        num_cols = sum(int(Emu().emulator_record_num_transducers(self._ptr, dev_idx)) for dev_idx in range(dev_num))
        series = np.zeros([num_cols], dtype=pl.Series)
        i = 0
        for dev_idx in range(dev_num):
            for tr_idx in range(int(Emu().emulator_record_num_transducers(self._ptr, dev_idx))):
                Emu().emulator_record_output_ultrasound(
                    self._ptr,
                    dev_idx,
                    tr_idx,
                    v.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),  # type: ignore[arg-type]
                )
                series[i] = pl.Series(name=f"p_{dev_idx}_{tr_idx}[a.u.]", values=v)
                i += 1

        return pl.DataFrame(
            {
                "time[25us/256]": time,
                **{s.name: s for s in series},
            },
        )

    def sound_field(self: Self, range_: Range, option: RecordOption) -> SoundField:
        return SoundField(
            _validate_ptr(
                Emu().emulator_wait_sound_field(
                    self._handle,
                    Emu().emulator_sound_field(self._ptr, range_._inner, option._inner),
                ),
            ),
            self._handle,
        )

    async def sound_field_async(self: Self, range_: Range, option: RecordOption) -> SoundField:
        future: asyncio.Future = asyncio.Future()
        loop = asyncio.get_event_loop()
        ffi_future = Emu().emulator_sound_field(self._ptr, range_._inner, option._inner)
        loop.call_soon(
            lambda *_: future.set_result(Emu().emulator_wait_sound_field(self._handle, ffi_future)),
        )
        return SoundField(
            _validate_ptr(await future),
            self._handle,
        )


class Emulator(Geometry):
    _ptr: EmulatorPtr
    _runtime: RuntimePtr
    _handle: HandlePtr

    def __new__(cls: type["Emulator"]) -> "Emulator":
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def __private_new__(cls: type["Emulator"], builder: _Builder) -> "Emulator":
        ptr = Emu().emulator(builder._ptr())
        geometry_ptr = Emu().emulator_geometry(ptr)
        ins = super().__new__(cls)
        ins._ptr = Emu().emulator(builder._ptr())
        ins._runtime = Base().create_runtime()
        ins._handle = Base().get_runtime_handle(ins._runtime)
        ins.__private_init__(geometry_ptr)
        return ins

    def __private_init__(self: Self, geometry_ptr: GeometryPtr) -> None:
        super().__init__(geometry_ptr)

    @property
    def geometry(self: Self) -> Geometry:
        return self

    def record(self: Self, f: Callable[[Controller[Recorder]], Controller[Recorder]]) -> Record:
        return self.record_from(timedelta(seconds=0), f)

    async def record_async(self: Self, f: Callable[[Controller[Recorder]], Controller[Recorder]]) -> Record:
        return await self.record_from_async(timedelta(seconds=0), f)

    def record_from(self: Self, t: timedelta, f: Callable[[Controller[Recorder]], Controller[Recorder]]) -> Record:
        def f_native(ptr: ControllerPtr) -> None:
            geometry = Base().geometry(ptr)
            link = Base().link_get(ptr)
            cnt = Controller(geometry, self._runtime, self._handle, ptr, Recorder(self._handle, link))
            cnt = f(cnt)
            cnt._disposed = True

        f_native_ = ctypes.CFUNCTYPE(None, ControllerPtr)(f_native)

        return Record(
            _validate_ptr(
                Emu().emulator_wait_result_record(
                    self._handle,
                    Emu().emulator_record_from(
                        self._ptr,
                        int(t.total_seconds() * 1000 * 1000 * 1000),
                        f_native_,  # type: ignore[arg-type]
                    ),
                ),
            ),
            self._handle,
        )

    async def record_from_async(self: Self, t: timedelta, f: Callable[[Controller[Recorder]], Controller[Recorder]]) -> Record:
        def f_native(ptr: ControllerPtr) -> None:
            geometry = Base().geometry(ptr)
            link = Base().link_get(ptr)
            cnt = Controller(geometry, self._runtime, self._handle, ptr, Recorder(self._handle, link))
            cnt = f(cnt)
            cnt._disposed = True

        f_native_ = ctypes.CFUNCTYPE(None, ControllerPtr)(f_native)
        future: asyncio.Future = asyncio.Future()
        loop = asyncio.get_event_loop()
        ffi_future = Emu().emulator_record_from(self._ptr, int(t.total_seconds() * 1000 * 1000 * 1000), f_native_)  # type: ignore[arg-type]
        loop.call_soon(
            lambda *_: future.set_result(Emu().emulator_wait_result_record(self._handle, ffi_future)),
        )
        return Record(
            _validate_ptr(await future),
            self._handle,
        )

    def __del__(self: Self) -> None:
        self._dispose()

    def _dispose(self: Self) -> None:
        if self._ptr._0 is not None:
            Emu().emulator_free(self._ptr)
            self._ptr._0 = None

    def __enter__(self: Self) -> Self:
        return self

    def __exit__(
        self: Self,
        _exc_type: type[BaseException] | None,
        _exc_value: BaseException | None,
        _traceback: TracebackType | None,
    ) -> None:
        self._dispose()


_Builder.into_emulator = lambda self: Emulator.__private_new__(self)  # type: ignore[method-assign]



from datetime import timedelta
from typing import Self

from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.driver.datagram.stm.foci import FociSTM
from pyautd3.driver.datagram.stm.gain import GainSTM
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, SilencerTarget


class FixedCompletionTime:
    intensity: timedelta
    phase: timedelta

    def __init__(self: Self, *, intensity: timedelta, phase: timedelta) -> None:
        self.intensity = intensity
        self.phase = phase

    def _is_valid(
        self: Self,
        v: Modulation | FociSTM | GainSTM,
        strict_mode: bool,  # noqa: FBT001
    ) -> bool:
        return bool(
            Base().datagram_silencer_fixed_completion_time_is_valid(
                int(self.intensity.total_seconds() * 1000 * 1000 * 1000),
                int(self.phase.total_seconds() * 1000 * 1000 * 1000),
                strict_mode,
                v._sampling_config_intensity()._inner,
                v._sampling_config_phase()._inner,
            ),
        )

    def _datagram_ptr(self: Self, strict_mode: bool, target: SilencerTarget) -> DatagramPtr:  # noqa: FBT001
        return Base().datagram_silencer_from_completion_time(
            int(self.intensity.total_seconds() * 1000 * 1000 * 1000),
            int(self.phase.total_seconds() * 1000 * 1000 * 1000),
            strict_mode,
            target,
        )



import ctypes
from collections.abc import Iterable
from datetime import timedelta
from typing import Generic
from typing import Self
from typing import TypeVar
import numpy as np
from numpy.typing import ArrayLike
from pyautd3.derive import datagram
from pyautd3.derive.derive_builder import builder
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.datagram.stm.control_point import ControlPoints1
from pyautd3.driver.datagram.stm.control_point import ControlPoints2
from pyautd3.driver.datagram.stm.control_point import ControlPoints3
from pyautd3.driver.datagram.stm.control_point import ControlPoints4
from pyautd3.driver.datagram.stm.control_point import ControlPoints5
from pyautd3.driver.datagram.stm.control_point import ControlPoints6
from pyautd3.driver.datagram.stm.control_point import ControlPoints7
from pyautd3.driver.datagram.stm.control_point import ControlPoints8
from pyautd3.driver.datagram.stm.control_point import IControlPoints
from pyautd3.driver.datagram.stm.stm_sampling_config import STMSamplingConfig
from pyautd3.driver.datagram.with_segment import DatagramS
from pyautd3.driver.defined.freq import Freq
from pyautd3.driver.firmware.fpga import LoopBehavior
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.driver.firmware.fpga.transition_mode import TransitionMode
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr
from pyautd3.native_methods.autd3capi_driver import FociSTMPtr
from pyautd3.native_methods.autd3capi_driver import Segment
from pyautd3.native_methods.autd3capi_driver import TransitionModeWrap
from pyautd3.native_methods.autd3capi_driver import LoopBehavior as _LoopBehavior
from pyautd3.native_methods.utils import _validate_ptr
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold
from pyautd3.native_methods.autd3capi_driver import Segment, TransitionModeWrap
from pyautd3.driver.datagram.with_segment import DatagramWithSegment

C = TypeVar("C", bound=IControlPoints)

class FociSTM(DatagramS[FociSTMPtr], Datagram, Generic[C]):
    _points: list[C]
    _stm_sampling_config: STMSamplingConfig
    def __private_init__(self: FociSTM[C], sampling_config: STMSamplingConfig, foci: list[ArrayLike] | list[ControlPoints1] | list[ControlPoints2] | list[ControlPoints3] | list[ControlPoints4] | list[ControlPoints5] | list[ControlPoints6] | list[ControlPoints7] | list[ControlPoints8]) -> None: ...
    def __init__(self: FociSTM, config: SamplingConfig | Freq[float] | timedelta, iterable: Iterable[ArrayLike] | Iterable[ControlPoints1] | Iterable[ControlPoints2] | Iterable[ControlPoints3] | Iterable[ControlPoints4] | Iterable[ControlPoints5] | Iterable[ControlPoints6] | Iterable[ControlPoints7] | Iterable[ControlPoints8]) -> None: ...
    def _raw_ptr(self: FociSTM[C], _: Geometry) -> FociSTMPtr: ...
    def _ptr(self: FociSTM[C], ) -> FociSTMPtr: ...
    def _datagram_ptr(self: FociSTM[C], geometry: Geometry) -> DatagramPtr: ...
    def _into_segment(self: FociSTM[C], ptr: FociSTMPtr, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramPtr: ...
    def _sampling_config_intensity(self: FociSTM[C], ) -> SamplingConfig: ...
    def _sampling_config_phase(self: FociSTM[C], ) -> SamplingConfig: ...
    def with_loop_behavior(self: FociSTM[C], loop_behavior: _LoopBehavior) -> FociSTM[C]: ...
    def with_timeout(self: FociSTM[C], timeout: timedelta | None) -> DatagramWithTimeout[FociSTM[C]]: ...
    def with_parallel_threshold(self: FociSTM[C], threshold: int | None) -> DatagramWithParallelThreshold[FociSTM[C]]: ...
    def with_segment(self: FociSTM[C], segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramWithSegment[FociSTM[C]]: ...
    @classmethod
    def nearest(cls, config: Freq[float] | timedelta, iterable: Iterable[ArrayLike] | Iterable[ControlPoints1] | Iterable[ControlPoints2] | Iterable[ControlPoints3] | Iterable[ControlPoints4] | Iterable[ControlPoints5] | Iterable[ControlPoints6] | Iterable[ControlPoints7] | Iterable[ControlPoints8]) -> FociSTM: ...
    @property
    def freq(self: FociSTM[C]) -> Freq[float]: ...
    @property
    def period(self: FociSTM[C]) -> timedelta: ...
    @property
    def sampling_config(self: FociSTM[C]) -> SamplingConfig: ...
    @property
    def loop_behavior(self: FociSTM[C]) -> _LoopBehavior: ...



from collections.abc import Callable
from ctypes import CFUNCTYPE, c_uint8, c_uint16, c_void_p
from threading import Lock
from typing import Self

from pyautd3.derive import datagram
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.geometry import Geometry
from pyautd3.driver.geometry.device import Device
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, GeometryPtr


@datagram
class PulseWidthEncoder(Datagram):
    _cache: dict[int, Callable[[int], int]]
    _lock: Lock

    def __init__(self: Self, f: Callable[[Device], Callable[[int], int]] | None = None) -> None:
        super().__init__()
        self._cache = {}
        self._lock = Lock()

        if f is None:
            self._f_native = None
        else:

            def f_native(_context: c_void_p, geometry_ptr: GeometryPtr, dev_idx: int, idx: int) -> int:
                if dev_idx not in self._cache:
                    with self._lock:
                        self._cache[dev_idx] = f(Device(dev_idx, geometry_ptr))
                return self._cache[dev_idx](idx)

            self._f_native = CFUNCTYPE(c_uint8, c_void_p, GeometryPtr, c_uint16, c_uint8)(f_native)

    def _datagram_ptr(self: Self, geometry: Geometry) -> DatagramPtr:
        return (
            Base().datagram_pulse_width_encoder_default()
            if self._f_native is None
            else Base().datagram_pulse_width_encoder(self._f_native, None, geometry._geometry_ptr)  # type: ignore[arg-type]
        )



from .bessel import Bessel
from .custom import Custom
from .focus import Focus
from .group import Group
from .null import Null
from .plane import Plane
from .uniform import Uniform

__all__ = ["Focus", "Bessel", "Plane", "Null", "Group", "Custom", "Uniform"]



import ctypes
from collections.abc import Iterable
from datetime import timedelta
from typing import Generic, Self, TypeVar

import numpy as np
from numpy.typing import ArrayLike

from pyautd3.derive import datagram
from pyautd3.derive.derive_builder import builder
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.datagram.stm.control_point import (
    ControlPoints1,
    ControlPoints2,
    ControlPoints3,
    ControlPoints4,
    ControlPoints5,
    ControlPoints6,
    ControlPoints7,
    ControlPoints8,
    IControlPoints,
)
from pyautd3.driver.datagram.stm.stm_sampling_config import STMSamplingConfig
from pyautd3.driver.datagram.with_segment import DatagramS
from pyautd3.driver.defined.freq import Freq
from pyautd3.driver.firmware.fpga import LoopBehavior
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.driver.firmware.fpga.transition_mode import TransitionMode
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, FociSTMPtr, Segment, TransitionModeWrap
from pyautd3.native_methods.autd3capi_driver import LoopBehavior as _LoopBehavior
from pyautd3.native_methods.utils import _validate_ptr

__all__ = []  # type: ignore[var-annotated]

C = TypeVar("C", bound=IControlPoints)


@builder
@datagram
@datagram_with_segment
class FociSTM(
    DatagramS[FociSTMPtr],
    Datagram,
    Generic[C],
):
    _points: list[C]

    _stm_sampling_config: STMSamplingConfig
    _param_loop_behavior: _LoopBehavior

    def __private_init__(
        self: "FociSTM",
        sampling_config: STMSamplingConfig,
        foci: (
            list[ArrayLike]
            | list[ControlPoints1]
            | list[ControlPoints2]
            | list[ControlPoints3]
            | list[ControlPoints4]
            | list[ControlPoints5]
            | list[ControlPoints6]
            | list[ControlPoints7]
            | list[ControlPoints8]
        ),
    ) -> None:
        match foci[0]:
            case (
                ControlPoints1()
                | ControlPoints2()
                | ControlPoints3()
                | ControlPoints4()
                | ControlPoints5()
                | ControlPoints6()
                | ControlPoints7()
                | ControlPoints8()
            ):
                self._points = foci
            case _:
                self._points = [ControlPoints1(p) for p in foci]

        self._stm_sampling_config = sampling_config

        self._param_loop_behavior = LoopBehavior.Infinite

    def __init__(
        self: "FociSTM",
        config: "SamplingConfig | Freq[float] | timedelta",
        iterable: (
            Iterable[ArrayLike]
            | Iterable[ControlPoints1]
            | Iterable[ControlPoints2]
            | Iterable[ControlPoints3]
            | Iterable[ControlPoints4]
            | Iterable[ControlPoints5]
            | Iterable[ControlPoints6]
            | Iterable[ControlPoints7]
            | Iterable[ControlPoints8]
        ),
    ) -> None:
        foci = list(iterable)
        self.__private_init__(STMSamplingConfig(config, len(foci)), foci)

    @classmethod
    def nearest(
        cls: type["FociSTM"],
        config: "Freq[float] | timedelta",
        iterable: (
            Iterable[ArrayLike]
            | Iterable[ControlPoints1]
            | Iterable[ControlPoints2]
            | Iterable[ControlPoints3]
            | Iterable[ControlPoints4]
            | Iterable[ControlPoints5]
            | Iterable[ControlPoints6]
            | Iterable[ControlPoints7]
            | Iterable[ControlPoints8]
        ),
    ) -> "FociSTM":
        ins = cls.__new__(cls)
        foci = list(iterable)
        ins.__private_init__(STMSamplingConfig._nearest(config, len(foci)), foci)
        return ins

    def _raw_ptr(self: Self, _: Geometry) -> FociSTMPtr:
        return self._ptr()

    def _ptr(self: Self) -> FociSTMPtr:
        n = self._points[0]._value()
        points = np.fromiter((np.void(p) for p in self._points), dtype=np.dtype((np.void, 4 + n * 16)))  # type: ignore[type-var,call-overload]
        return _validate_ptr(
            Base().stm_foci(
                self._stm_sampling_config._inner,
                points.ctypes.data_as(ctypes.c_void_p),  # type: ignore[arg-type]
                len(self._points),
                n,
                self._param_loop_behavior,
            ),
        )

    def _datagram_ptr(self: Self, geometry: Geometry) -> DatagramPtr:
        return Base().stm_foci_into_datagram(self._raw_ptr(geometry), self._points[0]._value())

    def _into_segment(self: Self, ptr: FociSTMPtr, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramPtr:
        return Base().stm_foci_into_datagram_with_segment(
            ptr,
            self._points[0]._value(),
            segment,
            transition_mode if transition_mode is not None else TransitionMode.NONE,
        )

    @property
    def freq(self: Self) -> Freq[float]:
        return self._stm_sampling_config.freq()

    @property
    def period(self: Self) -> timedelta:
        return self._stm_sampling_config.period()

    @property
    def sampling_config(self: Self) -> SamplingConfig:
        return self._stm_sampling_config.sampling_config()

    def _sampling_config_intensity(self: Self) -> SamplingConfig:
        return self.sampling_config

    def _sampling_config_phase(self: Self) -> SamplingConfig:
        return self.sampling_config



from ctypes import Array, c_float
from typing import Self

from pyautd3.gain.holo.backend import Backend
from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.native_methods.autd3capi_gain_holo import EmissionConstraintWrap
from pyautd3.native_methods.autd3capi_gain_holo import NativeMethods as GainHolo
from pyautd3.native_methods.structs import Vector3


class NalgebraBackend(Backend):
    def __init__(self: Self) -> None:
        super().__init__(GainHolo().nalgebra_backend_sphere())

    def __del__(self: Self) -> None:
        if self._ptr._0 is not None:
            GainHolo().delete_nalgebra_backend_sphere(self._ptr)
            self._ptr._0 = None

    def _gs(
        self: Self,
        foci: Array[Vector3],
        amps: Array[c_float],
        size: int,
        repeat: int,
        constraint: EmissionConstraintWrap,
    ) -> GainPtr:
        return GainHolo().gain_holo_gs_sphere(self._backend_ptr(), foci, amps, size, repeat, constraint)

    def _gspat(
        self: Self,
        foci: Array[Vector3],
        amps: Array[c_float],
        size: int,
        repeat: int,
        constraint: EmissionConstraintWrap,
    ) -> GainPtr:
        return GainHolo().gain_holo_gspat_sphere(self._backend_ptr(), foci, amps, size, repeat, constraint)

    def _naive(self: Self, foci: Array[Vector3], amps: Array[c_float], size: int, constraint: EmissionConstraintWrap) -> GainPtr:
        return GainHolo().gain_holo_naive_sphere(self._backend_ptr(), foci, amps, size, constraint)

    def _lm(
        self: Self,
        foci: Array[Vector3],
        amps: Array[c_float],
        size: int,
        eps1: float,
        eps2: float,
        tau: float,
        kmax: int,
        initial: Array[c_float],
        initial_size: int,
        constraint: EmissionConstraintWrap,
    ) -> GainPtr:
        return GainHolo().gain_holo_lm_sphere(
            self._backend_ptr(),
            foci,
            amps,
            size,
            eps1,
            eps2,
            tau,
            kmax,
            initial,
            initial_size,
            constraint,
        )



from typing import Self

from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.driver.datagram.stm.foci import FociSTM
from pyautd3.driver.datagram.stm.gain import GainSTM
from pyautd3.driver.utils import _validate_nonzero_u16
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, SilencerTarget


class FixedUpdateRate:
    intensity: int
    phase: int

    def __init__(self: Self, *, intensity: int, phase: int) -> None:
        self.intensity = _validate_nonzero_u16(intensity)
        self.phase = _validate_nonzero_u16(phase)

    def _is_valid(self: Self, v: Modulation | FociSTM | GainSTM, _strict_mode: bool) -> bool:  # noqa: FBT001
        return bool(
            Base().datagram_silencer_fixed_update_rate_is_valid(
                self.intensity,
                self.phase,
                v._sampling_config_intensity()._inner,
                v._sampling_config_phase()._inner,
            ),
        )

    def _datagram_ptr(self: Self, _strict_mode: bool, target: SilencerTarget) -> DatagramPtr:  # noqa: FBT001
        return Base().datagram_silencer_from_update_rate(
            self.intensity,
            self.phase,
            target,
        )



from datetime import timedelta
from pathlib import Path
from typing import Self

from pyautd3.derive import datagram, modulation
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.driver.defined.freq import Freq
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.modulation.resample import Resampler
from pyautd3.native_methods.autd3capi_driver import ModulationPtr
from pyautd3.native_methods.autd3capi_modulation_audio_file import NativeMethods as ModulationAudioFile
from pyautd3.native_methods.utils import _to_null_terminated_utf8, _validate_ptr


@modulation
@datagram_with_segment
@datagram
class RawPCM(Modulation):
    _path: Path
    _config: SamplingConfig | tuple[Freq[float], SamplingConfig, Resampler]
    _sample_rate: Freq[int]

    def __private_init__(self: Self, path: Path, config: SamplingConfig | tuple[Freq[float], SamplingConfig, Resampler]) -> None:
        super().__init__()
        self._path = path
        self._config = config

    def __init__(self: Self, path: Path, config: SamplingConfig | Freq[int] | Freq[float] | timedelta) -> None:
        self.__private_init__(path, SamplingConfig(config))

    @staticmethod
    def new_with_resample(
        path: Path,
        source: Freq[float],
        target: SamplingConfig | Freq[int] | Freq[float] | timedelta,
        resampler: Resampler,
    ) -> "RawPCM":
        instance = super(RawPCM, RawPCM).__new__(RawPCM)
        instance.__private_init__(path, (source, SamplingConfig(target), resampler))
        return instance

    def _modulation_ptr(self: Self) -> ModulationPtr:
        path = _to_null_terminated_utf8(str(self._path))
        match self._config:
            case (Freq(), SamplingConfig(), Resampler()):
                (source, target, resampler) = self._config
                return _validate_ptr(
                    ModulationAudioFile().modulation_audio_file_raw_pcm_with_resample(
                        path,
                        self._loop_behavior,
                        source.hz,
                        target._inner,
                        resampler._dyn_resampler(),
                    ),
                )
            case _:
                return _validate_ptr(
                    ModulationAudioFile().modulation_audio_file_raw_pcm(
                        _to_null_terminated_utf8(str(self._path)),
                        self._config._inner,
                        self._loop_behavior,
                    ),
                )



from .custom import Custom
from .fourier import Fourier
from .resample import BlackMan, Rectangular, Resampler, SincInterpolation
from .sine import Sine
from .square import Square
from .static import Static

__all__ = [
    "Static",
    "Sine",
    "Fourier",
    "Square",
    "Custom",
    "Resampler",
    "BlackMan",
    "Rectangular",
    "SincInterpolation",
]



import numpy as np
import pytest

from pyautd3 import AUTD3, Controller, EulerAngles, deg, rad
from pyautd3.driver.geometry.rotation import Angle
from pyautd3.link.audit import Audit

from .test_autd import create_controller


def test_angle():
    assert (np.pi / 2 * rad).radian == np.pi / 2
    assert (90 * deg).radian == np.pi / 2


def test_angle_ctr():
    with pytest.raises(NotImplementedError):
        _ = Angle()

    with pytest.raises(NotImplementedError):
        _ = Angle._UnitDegree()

    with pytest.raises(NotImplementedError):
        _ = Angle._UnitRad()

    with pytest.raises(NotImplementedError):
        _ = EulerAngles()


def test_with_rotation_xyz():
    def open_with_rotation(q: np.ndarray) -> Controller[Audit]:
        return Controller.builder([AUTD3([0.0, 0.0, 0.0]).with_rotation(q)]).open(Audit.builder())

    with open_with_rotation(EulerAngles.XYZ(90 * deg, 0 * deg, 0 * deg)) as autd:
        assert np.allclose(autd.geometry[0].x_direction, [1.0, 0.0, 0.0])
        assert np.allclose(autd.geometry[0].y_direction, [0.0, 0.0, 1.0])
        assert np.allclose(autd.geometry[0].axial_direction, [0.0, -1.0, 0.0])

    with open_with_rotation(EulerAngles.XYZ(0 * deg, 90 * deg, 0 * deg)) as autd:
        assert np.allclose(autd.geometry[0].x_direction, [0.0, 0.0, -1.0])
        assert np.allclose(autd.geometry[0].y_direction, [0.0, 1.0, 0.0])
        assert np.allclose(autd.geometry[0].axial_direction, [1.0, 0.0, 0.0])

    with open_with_rotation(EulerAngles.XYZ(0 * deg, 0 * deg, 90 * deg)) as autd:
        assert np.allclose(autd.geometry[0].x_direction, [0.0, 1.0, 0.0])
        assert np.allclose(autd.geometry[0].y_direction, [-1.0, 0.0, 0.0])
        assert np.allclose(autd.geometry[0].axial_direction, [0.0, 0.0, 1.0])

    with open_with_rotation(EulerAngles.XYZ(0 * deg, 90 * deg, 90 * deg)) as autd:
        assert np.allclose(autd.geometry[0].x_direction, [0.0, 1.0, 0.0])
        assert np.allclose(autd.geometry[0].y_direction, [0.0, 0.0, 1.0])
        assert np.allclose(autd.geometry[0].axial_direction, [1.0, 0.0, 0.0])

    with open_with_rotation(EulerAngles.XYZ(90 * deg, 90 * deg, 0 * deg)) as autd:
        assert np.allclose(autd.geometry[0].x_direction, [0.0, 1.0, 0.0])
        assert np.allclose(autd.geometry[0].y_direction, [0.0, 0.0, 1.0])
        assert np.allclose(autd.geometry[0].axial_direction, [1.0, 0.0, 0.0])


def test_with_rotation_zyz():
    def open_with_rotation(q: np.ndarray) -> Controller[Audit]:
        return Controller.builder([AUTD3([0.0, 0.0, 0.0]).with_rotation(q)]).open(Audit.builder())

    with open_with_rotation(EulerAngles.ZYZ(90 * deg, 0 * deg, 0 * deg)) as autd:
        assert np.allclose(autd.geometry[0].x_direction, [0.0, 1.0, 0.0])
        assert np.allclose(autd.geometry[0].y_direction, [-1.0, 0.0, 0.0])
        assert np.allclose(autd.geometry[0].axial_direction, [0.0, 0.0, 1.0])

    with open_with_rotation(EulerAngles.ZYZ(0 * deg, 90 * deg, 0 * deg)) as autd:
        assert np.allclose(autd.geometry[0].x_direction, [0.0, 0.0, -1.0])
        assert np.allclose(autd.geometry[0].y_direction, [0.0, 1.0, 0.0])
        assert np.allclose(autd.geometry[0].axial_direction, [1.0, 0.0, 0.0])

    with open_with_rotation(EulerAngles.ZYZ(0 * deg, 0 * deg, 90 * deg)) as autd:
        assert np.allclose(autd.geometry[0].x_direction, [0.0, 1.0, 0.0])
        assert np.allclose(autd.geometry[0].y_direction, [-1.0, 0.0, 0.0])
        assert np.allclose(autd.geometry[0].axial_direction, [0.0, 0.0, 1.0])

    with open_with_rotation(EulerAngles.ZYZ(0 * deg, 90 * deg, 90 * deg)) as autd:
        assert np.allclose(autd.geometry[0].x_direction, [0.0, 1.0, 0.0])
        assert np.allclose(autd.geometry[0].y_direction, [0.0, 0.0, 1.0])
        assert np.allclose(autd.geometry[0].axial_direction, [1.0, 0.0, 0.0])

    with open_with_rotation(EulerAngles.ZYZ(90 * deg, 90 * deg, 0 * deg)) as autd:
        assert np.allclose(autd.geometry[0].x_direction, [0.0, 0.0, -1.0])
        assert np.allclose(autd.geometry[0].y_direction, [-1.0, 0.0, 0.0])
        assert np.allclose(autd.geometry[0].axial_direction, [0.0, 1.0, 0.0])


def test_geometry_num_devices():
    with create_controller() as autd:
        assert autd.num_devices == 2


def test_geometry_center():
    with create_controller() as autd:
        center = autd.center
        assert len(center) == 3
        assert center[0] == 86.625267028808594
        assert center[1] == 66.71319580078125
        assert center[2] == 0.0


def test_device_idx():
    with create_controller() as autd:
        assert autd.geometry[0].idx == 0
        assert autd.geometry[1].idx == 1


def test_device_sound_speed():
    with create_controller() as autd:
        for dev in autd.geometry:
            assert dev.sound_speed == 340e3
            dev.sound_speed = 350e3
            assert dev.sound_speed == 350e3


def test_device_set_sound_speed_from_temp():
    with create_controller() as autd:
        for dev in autd.geometry:
            dev.set_sound_speed_from_temp(15)
            assert dev.sound_speed == 340.29525e3

        autd.set_sound_speed(350e3)
        for dev in autd.geometry:
            assert dev.sound_speed == 350e3

        autd.set_sound_speed_from_temp(15)
        for dev in autd.geometry:
            assert dev.sound_speed == 340.29525e3


def test_device_enable():
    with create_controller() as autd:
        for dev in autd.geometry:
            assert dev.enable
            dev.enable = False
            assert not dev.enable


def test_device_num_transducers():
    with create_controller() as autd:
        assert autd.num_transducers == 249 * autd.num_devices
        for dev in autd.geometry:
            assert dev.num_transducers == 249


def test_device_center():
    with create_controller() as autd:
        for dev in autd.geometry:
            center = dev.center
            assert len(center) == 3
            assert center[0] == 86.625267028808594
            assert center[1] == 66.71319580078125
            assert center[2] == 0.0


def test_device_translate():
    with create_controller() as autd:
        for dev in autd.geometry:
            original_pos = [tr.position for tr in dev]
            t = [1, 2, 3]
            dev.translate(t)
            for tr in dev:
                assert np.allclose(tr.position, original_pos[tr.idx] + t)


def test_device_rotate():
    with create_controller() as autd:
        for dev in autd.geometry:
            r = [0.70710678, 0.0, 0.0, 0.70710678]
            dev.rotate(r)
            assert np.allclose(dev.rotation, r)


def test_device_affine():
    with create_controller() as autd:
        for dev in autd.geometry:
            original_pos = [tr.position for tr in dev]
            t = [1, 2, 3]
            r = [0.70710678, 0.0, 0.0, 0.70710678]
            dev.affine(t, r)
            for tr in dev:
                op = original_pos[tr.idx]
                expected = np.array([-op[1], op[0], op[2]]) + t
                assert np.allclose(tr.position, expected)
            assert np.allclose(dev.rotation, r)


def test_device_wavelength():
    with create_controller() as autd:
        for dev in autd.geometry:
            assert dev.wavelength == 340e3 / 40e3


def test_device_wavenum():
    with create_controller() as autd:
        for dev in autd.geometry:
            assert dev.wavenumber == 0.7391983270645142


def test_transducer_idx():
    with create_controller() as autd:
        for dev in autd.geometry:
            for i, tr in enumerate(dev):
                assert tr.idx == i


def test_transducer_dev_idx():
    with create_controller() as autd:
        for dev in autd.geometry:
            for tr in dev:
                assert tr.dev_idx == dev.idx


def test_transducer_position():
    with create_controller() as autd:
        assert np.allclose(autd.geometry[0][0].position, [0.0, 0.0, 0.0])
        assert np.allclose(
            autd.geometry[0][-1].position,
            [(AUTD3.NUM_TRANS_IN_X - 1) * AUTD3.TRANS_SPACING, (AUTD3.NUM_TRANS_IN_Y - 1) * AUTD3.TRANS_SPACING, 0.0],
        )

        assert np.allclose(autd.geometry[1][0].position, [0.0, 0.0, 0.0])
        assert np.allclose(
            autd.geometry[1][-1].position,
            [(AUTD3.NUM_TRANS_IN_X - 1) * AUTD3.TRANS_SPACING, (AUTD3.NUM_TRANS_IN_Y - 1) * AUTD3.TRANS_SPACING, 0.0],
        )


def test_transducer_rotation():
    with create_controller() as autd:
        for dev in autd.geometry:
            assert np.allclose(dev.rotation, [1.0, 0.0, 0.0, 0.0])


def test_transducer_x_direction():
    with create_controller() as autd:
        for dev in autd.geometry:
            assert np.allclose(dev.x_direction, [1.0, 0.0, 0.0])


def test_transducer_y_direction():
    with create_controller() as autd:
        for dev in autd.geometry:
            assert np.allclose(dev.y_direction, [0.0, 1.0, 0.0])


def test_transducer_axial_direction():
    with create_controller() as autd:
        for dev in autd.geometry:
            assert np.allclose(dev.axial_direction, [0.0, 0.0, 1.0])



from typing import Self, TypeVar

from pyautd3.derive import builder, datagram, modulation
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.modulation import ModulationWithSamplingConfig
from pyautd3.driver.defined.angle import Angle, rad
from pyautd3.driver.defined.freq import Freq, Hz
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.modulation.sampling_mode import ISamplingMode, SamplingModeExact, SamplingModeExactFloat, SamplingModeNearest
from pyautd3.native_methods.autd3capi_driver import ModulationPtr

T = TypeVar("T", int, float)


@modulation
@datagram
@datagram_with_segment
@builder
class Sine(ModulationWithSamplingConfig):
    _mode: ISamplingMode
    _param_intensity_u8: int
    _param_offset_u8: int
    _param_phase: Angle
    _param_clamp: bool

    def __private__init__(self: Self, mode: ISamplingMode) -> None:
        super().__init__(SamplingConfig(10))
        self._mode = mode
        self._param_intensity_u8 = 0xFF
        self._param_offset_u8 = 0x80
        self._param_phase = 0 * rad
        self._param_clamp = False

    def __init__(self: Self, freq: Freq[T]) -> None:
        match freq.hz:
            case int():
                self.__private__init__(SamplingModeExact(freq))  # type: ignore[arg-type]
            case _:
                self.__private__init__(SamplingModeExactFloat(freq))

    @classmethod
    def nearest(cls: "type[Sine]", freq: Freq[float]) -> "Sine":
        sine = super().__new__(cls)
        sine.__private__init__(SamplingModeNearest(freq))
        return sine

    @property
    def freq(self: Self) -> Freq[int] | Freq[float]:
        return self._mode.sine_freq() * Hz

    def _modulation_ptr(self: Self) -> ModulationPtr:
        return self._mode.sine_ptr(
            self._config._inner,
            self._param_intensity_u8,
            self._param_offset_u8,
            self._param_phase,
            self._param_clamp,
            self._loop_behavior,
        )



from typing import Self

import numpy as np


class Angle:
    _value: float

    def __new__(cls: type["Angle"]) -> "Angle":
        raise NotImplementedError

    @classmethod
    def __private_new__(cls: type["Angle"], value: float) -> "Angle":
        ins = super().__new__(cls)
        ins._value = value
        return ins

    @property
    def radian(self: Self) -> float:
        return self._value

    def __eq__(self: Self, other: object) -> bool:
        return isinstance(other, Angle) and self._value == other._value

    class _UnitRad:
        def __new__(cls: type["Angle._UnitRad"]) -> "Angle._UnitRad":
            raise NotImplementedError

        @classmethod
        def __private_new__(cls: type["Angle._UnitRad"]) -> "Angle._UnitRad":
            return super().__new__(cls)

        def __rmul__(self: Self, other: float) -> "Angle":
            return Angle.__private_new__(other)

    class _UnitDegree:
        def __new__(cls: type["Angle._UnitDegree"]) -> "Angle._UnitDegree":
            raise NotImplementedError

        @classmethod
        def __private_new__(cls: type["Angle._UnitDegree"]) -> "Angle._UnitDegree":
            return super().__new__(cls)

        def __rmul__(self: Self, other: float) -> "Angle":
            return Angle.__private_new__(np.deg2rad(other))


rad: Angle._UnitRad = Angle._UnitRad.__private_new__()
deg: Angle._UnitDegree = Angle._UnitDegree.__private_new__()



from collections.abc import Callable
from ctypes import POINTER
from ctypes import c_int32
from ctypes import c_uint16
from typing import Generic
from typing import Self
from typing import TypeVar
import numpy as np
from pyautd3.autd_error import UnknownGroupKeyError
from pyautd3.derive import datagram
from pyautd3.derive import gain
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.gain import Gain
from pyautd3.driver.geometry import Device
from pyautd3.driver.geometry import Geometry
from pyautd3.driver.geometry import Transducer
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.gain.cache import Cache
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold
from pyautd3.native_methods.autd3capi_driver import Segment, TransitionModeWrap
from pyautd3.driver.datagram.with_segment import DatagramWithSegment

K = TypeVar("K")

class Group(Gain, Generic[K]):
    _map: dict[K, Gain]
    _f: Callable[[Device], Callable[[Transducer], K | None]]
    def __init__(self: Group[K], f: Callable[[Device], Callable[[Transducer], K | None]]) -> None: ...
    def set(self: Group[K], key: K, gain: Gain) -> Group[K]: ...
    def _gain_ptr(self: Group[K], geometry: Geometry) -> GainPtr: ...
    def with_cache(self: Group[K], ) -> Cache[Group[K]]: ...
    def with_timeout(self: Group[K], timeout: timedelta | None) -> DatagramWithTimeout[Group[K]]: ...
    def with_parallel_threshold(self: Group[K], threshold: int | None) -> DatagramWithParallelThreshold[Group[K]]: ...
    def with_segment(self: Group[K], segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramWithSegment[Group[K]]: ...



from datetime import timedelta
from typing import Generic, Self, TypeVar

from pyautd3.derive import datagram
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr

__all__ = []  # type: ignore[var-annotated]

D = TypeVar("D", bound="Datagram")


@datagram
class DatagramWithTimeout(Datagram, Generic[D]):
    _datagram: D
    _timeout: timedelta | None

    def __init__(self: Self, datagram: D, timeout: timedelta | None) -> None:
        self._datagram = datagram
        self._timeout = timeout

    def _datagram_ptr(self: Self, g: Geometry) -> DatagramPtr:
        raw_ptr = self._datagram._datagram_ptr(g)
        return Base().datagram_with_timeout(raw_ptr, int(self._timeout.total_seconds() * 1000 * 1000 * 1000) if self._timeout is not None else -1)



import ctypes
from abc import ABCMeta, abstractmethod
from typing import Generic, Self, TypeVar

from pyautd3.native_methods.autd3capi_driver import DynSincInterpolator, DynWindow


class Resampler(metaclass=ABCMeta):
    @abstractmethod
    def _dyn_resampler(self: Self) -> DynSincInterpolator:
        pass


class BlackMan:
    _window_size: int

    def __init__(self: Self, window_size: int) -> None:
        self._window_size = window_size

    def _window(self: Self) -> ctypes.c_uint32:
        return ctypes.c_uint32(int(DynWindow.Blackman))


class Rectangular:
    _window_size: int

    def __init__(self: Self, window_size: int) -> None:
        self._window_size = window_size

    def _window(self: Self) -> ctypes.c_uint32:
        return ctypes.c_uint32(int(DynWindow.Rectangular))


T = TypeVar("T", BlackMan, Rectangular)


class SincInterpolation(
    Resampler,
    Generic[T],
):
    _window: T

    def __init__(self: "SincInterpolation", window: T | None = None) -> None:
        self._window = window if window is not None else BlackMan(32)

    def _dyn_resampler(self: Self) -> DynSincInterpolator:
        return DynSincInterpolator(self._window._window(), ctypes.c_uint32(self._window._window_size))



import ctypes
from typing import Self

from pyautd3.native_methods.autd3capi import NativeMethods as Base


class FirmwareInfo:
    _info: str

    def __init__(self: Self, info: str) -> None:
        self._info = info

    @property
    def info(self: Self) -> str:
        return self._info

    @staticmethod
    def latest_version() -> str:
        sb = ctypes.create_string_buffer(256)
        Base().firmware_latest(sb)
        return sb.value.decode("utf-8")

    def __str__(self: Self) -> str:
        return self._info



from datetime import timedelta
from typing import Generic, Self, TypeVar

from pyautd3.derive import builder, datagram
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.driver.datagram.silencer.fixed_completion_time import FixedCompletionTime
from pyautd3.driver.datagram.silencer.fixed_update_rate import FixedUpdateRate
from pyautd3.driver.datagram.stm.foci import FociSTM
from pyautd3.driver.datagram.stm.gain import GainSTM
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, SilencerTarget

T = TypeVar("T", FixedCompletionTime, FixedUpdateRate)


@builder
@datagram
class Silencer(
    Datagram,
    Generic[T],
):
    _inner: T
    _strict_mode: bool
    _param_target: SilencerTarget

    def __init__(self: Self, config: T | None = None) -> None:
        super().__init__()
        self._inner = (
            config
            if config is not None
            else FixedCompletionTime(
                intensity=timedelta(microseconds=250),
                phase=timedelta(microseconds=1000),
            )  # type: ignore[assignment]
        )
        self._strict_mode = True
        self._param_target = SilencerTarget.Intensity

    def with_strict_mode(self: "Silencer[T]", mode: bool) -> "Silencer[T]":  # noqa: FBT001
        if not isinstance(self._inner, FixedCompletionTime):  # pragma: no cover
            msg = "Strict mode is only available for Silencer[FixedCompletionTime]"  # pragma: no cover
            raise TypeError(msg)  # pragma: no cover
        self._strict_mode = mode
        return self

    def is_valid(self: Self, target: Modulation | FociSTM | GainSTM) -> bool:
        return self._inner._is_valid(target, self._strict_mode)

    def _datagram_ptr(self: Self, _: Geometry) -> DatagramPtr:
        return self._inner._datagram_ptr(self._strict_mode, self._param_target)

    @staticmethod
    def disable() -> "Silencer[FixedCompletionTime]":
        return Silencer(FixedCompletionTime(intensity=timedelta(microseconds=25), phase=timedelta(microseconds=25)))



from typing import Self

from pyautd3.driver.utils import _validate_u8


class EmitIntensity:
    _value: int

    def __init__(self: Self, intensity: "int | EmitIntensity") -> None:
        match intensity:
            case EmitIntensity():
                self._value = intensity._value
            case int():
                self._value = _validate_u8(intensity)
            case _:
                raise TypeError

    @property
    def value(self: Self) -> int:
        return self._value

    @staticmethod
    def maximum() -> "EmitIntensity":
        return EmitIntensity(0xFF)

    @staticmethod
    def minimum() -> "EmitIntensity":
        return EmitIntensity(0x00)

    def __floordiv__(self: Self, other: int) -> "EmitIntensity":
        return EmitIntensity(self.value // other)

    def __eq__(self: Self, __value: object) -> bool:
        return isinstance(__value, EmitIntensity) and self.value == __value.value

    def __str__(self: Self) -> str:
        return f"EmitIntensity({self.value})"

    def __repr__(self: Self) -> str:
        return f"EmitIntensity({self.value})"



from datetime import timedelta

import numpy as np
import pytest

from pyautd3 import AUTD3, Clear, Controller, Device, ForceFan, Segment, tracing_init
from pyautd3.autd_error import AUTDError, InvalidDatagramTypeError, KeyAlreadyExistsError
from pyautd3.controller.timer import AsyncSleeper, SpinSleeper, StdSleeper, TimerStrategy, WaitableSleeper
from pyautd3.driver.datagram import Synchronize
from pyautd3.driver.defined.freq import Hz
from pyautd3.driver.firmware.fpga.emit_intensity import EmitIntensity
from pyautd3.driver.firmware.fpga.phase import Phase
from pyautd3.driver.firmware_version import FirmwareInfo
from pyautd3.gain import Null, Uniform
from pyautd3.link.audit import Audit
from pyautd3.modulation import Sine, Static
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import SpinStrategyTag


async def create_controller_async() -> Controller[Audit]:
    return (
        await Controller.builder([AUTD3([0.0, 0.0, 0.0]), AUTD3([0.0, 0.0, 0.0])])
        .with_send_interval(timedelta(milliseconds=1))
        .with_receive_interval(timedelta(milliseconds=1))
        .with_fallback_parallel_threshold(4)
        .with_fallback_timeout(timedelta(milliseconds=20))
        .with_timer_strategy(TimerStrategy.Spin(SpinSleeper()))
        .open_async(
            Audit.builder(),
        )
    )


def create_controller() -> Controller[Audit]:
    return (
        Controller.builder([AUTD3([0.0, 0.0, 0.0]), AUTD3([0.0, 0.0, 0.0])])
        .with_send_interval(timedelta(milliseconds=1))
        .with_receive_interval(timedelta(milliseconds=1))
        .with_fallback_parallel_threshold(4)
        .with_fallback_timeout(timedelta(milliseconds=20))
        .with_timer_strategy(TimerStrategy.Spin(SpinSleeper()))
        .open(Audit.builder())
    )


def test_controller_is_default():
    default = Controller.builder([])
    assert Base().controller_builder_is_default(
        default.fallback_parallel_threshold,
        int(default.fallback_timeout.total_seconds() * 1000 * 1000 * 1000),
        int(default.send_interval.total_seconds() * 1000 * 1000 * 1000),
        int(default.receive_interval.total_seconds() * 1000 * 1000 * 1000),
        default.timer_strategy,
    )


def test_with_timer():
    autd: Controller[Audit]
    with Controller.builder([]).with_timer_strategy(TimerStrategy.Std(StdSleeper())).open(Audit.builder()) as autd:
        autd.send(Static())
    with (
        Controller.builder([])
        .with_timer_strategy(
            TimerStrategy.Spin(SpinSleeper(700_000).with_spin_strategy(SpinStrategyTag.SpinLoopHint)),
        )
        .open(Audit.builder()) as autd
    ):
        autd.send(Static())
    with Controller.builder([]).with_timer_strategy(TimerStrategy.Async(AsyncSleeper())).open(Audit.builder()) as autd:
        autd.send(Static())

    _ = TimerStrategy.Waitable(WaitableSleeper())

    with pytest.raises(NotImplementedError):
        _ = TimerStrategy()


def test_firmware_info():
    autd: Controller[Audit]
    with create_controller() as autd:
        for i, firm in enumerate(autd.firmware_version()):
            assert firm.info == f"{i}: CPU = v10.0.1, FPGA = v10.0.1 [Emulator]"
            assert str(firm) == f"{i}: CPU = v10.0.1, FPGA = v10.0.1 [Emulator]"

        autd.link.down()
        with pytest.raises(AUTDError) as e:
            autd.firmware_version()
        assert str(e.value) == "Read firmware info failed: 0, 1"
        autd.link.up()


@pytest.mark.asyncio
async def test_firmware_info_async():
    autd: Controller[Audit]
    with await create_controller_async() as autd:
        assert FirmwareInfo.latest_version() == "v10.0.1"

        for i, firm in enumerate(await autd.firmware_version_async()):
            assert firm.info == f"{i}: CPU = v10.0.1, FPGA = v10.0.1 [Emulator]"
            assert str(firm) == f"{i}: CPU = v10.0.1, FPGA = v10.0.1 [Emulator]"

        autd.link.down()
        with pytest.raises(AUTDError) as e:
            await autd.firmware_version_async()
        assert str(e.value) == "Read firmware info failed: 0, 1"
        autd.link.up()


def test_close():
    autd: Controller[Audit]
    with create_controller() as autd:
        assert autd.link.is_open()

        autd.close()
        autd.close()

    with create_controller() as autd:
        autd.link.break_down()
        with pytest.raises(AUTDError) as e:
            autd.close()
        assert str(e.value) == "broken"


@pytest.mark.asyncio
async def test_close_async():
    autd: Controller[Audit]
    with await create_controller_async() as autd:
        assert autd.link.is_open()

        await autd.close_async()
        await autd.close_async()

    with create_controller() as autd:
        autd.link.break_down()
        with pytest.raises(AUTDError) as e:
            await autd.close_async()
        assert str(e.value) == "broken"


def test_send_single():
    autd: Controller[Audit]
    with create_controller() as autd:
        for dev in autd.geometry:
            assert np.all(autd.link.modulation_buffer(dev.idx, Segment.S0) == 0xFF)

        autd.send(Static())

        for dev in autd.geometry:
            assert np.all(autd.link.modulation_buffer(dev.idx, Segment.S0) == 0xFF)

        autd.link.down()
        with pytest.raises(AUTDError) as e:
            autd.send(Static())
        assert str(e.value) == "Failed to send data"
        autd.link.up()

        autd.link.break_down()
        with pytest.raises(AUTDError) as e:
            autd.send(Static())
        assert str(e.value) == "broken"
        autd.link.repair()


@pytest.mark.asyncio
async def test_send_async_single():
    autd: Controller[Audit]
    with await create_controller_async() as autd:
        for dev in autd.geometry:
            assert np.all(autd.link.modulation_buffer(dev.idx, Segment.S0) == 0xFF)

        await autd.send_async(Static())

        for dev in autd.geometry:
            assert np.all(autd.link.modulation_buffer(dev.idx, Segment.S0) == 0xFF)

        autd.link.down()
        with pytest.raises(AUTDError) as e:
            await autd.send_async(Static())
        assert str(e.value) == "Failed to send data"
        autd.link.up()

        autd.link.break_down()
        with pytest.raises(AUTDError) as e:
            await autd.send_async(Static())
        assert str(e.value) == "broken"
        autd.link.repair()


@pytest.mark.asyncio
async def test_send_async_tuple():
    autd: Controller[Audit]
    with await create_controller_async() as autd:
        for dev in autd.geometry:
            assert np.all(autd.link.modulation_buffer(dev.idx, Segment.S0) == 0xFF)
            intensities, phases = autd.link.drives_at(dev.idx, Segment.S0, 0)
            assert np.all(intensities == 0)
            assert np.all(phases == 0)

        await autd.send_async((Static(), Uniform(EmitIntensity(0x80))))
        for dev in autd.geometry:
            assert np.all(autd.link.modulation_buffer(dev.idx, Segment.S0) == 0xFF)
            intensities, phases = autd.link.drives_at(dev.idx, Segment.S0, 0)
            assert np.all(intensities == 0x80)
            assert np.all(phases == 0)

        with pytest.raises(InvalidDatagramTypeError):
            await autd.send_async(0)  # type: ignore[arg-type]

        autd.link.down()
        with pytest.raises(AUTDError) as e:
            await autd.send_async((Static(), Uniform(EmitIntensity(0xFF))))
        assert str(e.value) == "Failed to send data"
        autd.link.up()

        autd.link.break_down()
        with pytest.raises(AUTDError) as e:
            await autd.send_async((Static(), Uniform(EmitIntensity(0xFF))))
        assert str(e.value) == "broken"
        autd.link.repair()


def test_send_tuple():
    autd: Controller[Audit]
    with create_controller() as autd:
        for dev in autd.geometry:
            assert np.all(autd.link.modulation_buffer(dev.idx, Segment.S0) == 0xFF)
            intensities, phases = autd.link.drives_at(dev.idx, Segment.S0, 0)
            assert np.all(intensities == 0)
            assert np.all(phases == 0)

        autd.send((Static(), Uniform(EmitIntensity(0x80))))
        for dev in autd.geometry:
            assert np.all(autd.link.modulation_buffer(dev.idx, Segment.S0) == 0xFF)
            intensities, phases = autd.link.drives_at(dev.idx, Segment.S0, 0)
            assert np.all(intensities == 0x80)
            assert np.all(phases == 0)

        with pytest.raises(InvalidDatagramTypeError):
            autd.send(0)  # type: ignore[arg-type]

        autd.link.down()
        with pytest.raises(AUTDError) as e:
            autd.send((Static(), Uniform(EmitIntensity(0xFF))))
        assert str(e.value) == "Failed to send data"
        autd.link.up()

        autd.link.break_down()
        with pytest.raises(AUTDError) as e:
            autd.send((Static(), Uniform(EmitIntensity(0xFF))))
        assert str(e.value) == "broken"
        autd.link.repair()


@pytest.mark.asyncio
async def test_group_async():
    autd: Controller[Audit]
    with await create_controller_async() as autd:
        await autd.group(lambda dev: dev.idx).set(1, Null()).set(0, (Sine(150 * Hz), Uniform(EmitIntensity(0xFF)))).send_async()

        mod = autd.link.modulation_buffer(0, Segment.S0)
        assert len(mod) == 80
        intensities, phases = autd.link.drives_at(0, Segment.S0, 0)
        assert np.all(intensities == 0xFF)
        assert np.all(phases == 0)

        mod = autd.link.modulation_buffer(1, Segment.S0)
        intensities, phases = autd.link.drives_at(1, Segment.S0, 0)
        assert np.all(intensities == 0)

        with pytest.raises(InvalidDatagramTypeError):
            await autd.group(lambda dev: dev.idx).set(0, 0).send_async()  # type: ignore[arg-type]

        with pytest.raises(KeyAlreadyExistsError):
            await autd.group(lambda dev: dev.idx).set(0, Null()).set(0, Null()).send_async()


def test_group():
    autd: Controller[Audit]
    with create_controller() as autd:
        autd.group(lambda dev: dev.idx).set(1, Null()).set(0, (Sine(150 * Hz), Uniform(EmitIntensity(0xFF)))).send()

        mod = autd.link.modulation_buffer(0, Segment.S0)
        assert len(mod) == 80
        intensities, phases = autd.link.drives_at(0, Segment.S0, 0)
        assert np.all(intensities == 0xFF)
        assert np.all(phases == 0)

        mod = autd.link.modulation_buffer(1, Segment.S0)
        intensities, phases = autd.link.drives_at(1, Segment.S0, 0)
        assert np.all(intensities == 0)

        with pytest.raises(InvalidDatagramTypeError):
            autd.group(lambda dev: dev.idx).set(0, 0).send()  # type: ignore[arg-type]

        with pytest.raises(KeyAlreadyExistsError):
            autd.group(lambda dev: dev.idx).set(0, Null()).set(0, Null()).send()


def test_group_check_only_for_enabled():
    autd: Controller[Audit]
    with create_controller() as autd:
        check = np.zeros(autd.num_devices, dtype=bool)

        autd.geometry[0].enable = False

        def f(dev: Device) -> int:
            check[dev.idx] = True
            return 0

        autd.group(f).set(0, (Sine(150 * Hz), Uniform((EmitIntensity(0x80), Phase(0x90))))).send()

        assert not check[0]
        assert check[1]

        mod = autd.link.modulation_buffer(0, Segment.S0)
        intensities, phases = autd.link.drives_at(0, Segment.S0, 0)
        assert np.all(intensities == 0)
        assert np.all(phases == 0)

        mod = autd.link.modulation_buffer(1, Segment.S0)
        assert len(mod) == 80
        intensities, phases = autd.link.drives_at(1, Segment.S0, 0)
        assert np.all(intensities == 0x80)
        assert np.all(phases == 0x90)


def test_clear():
    autd: Controller[Audit]
    with create_controller() as autd:
        autd.send((Static(), Uniform((EmitIntensity(0xFF), Phase(0x90)))))

        for dev in autd.geometry:
            assert np.all(autd.link.modulation_buffer(dev.idx, Segment.S0) == 0xFF)
            intensities, phases = autd.link.drives_at(dev.idx, Segment.S0, 0)
            assert np.all(intensities == 0xFF)
            assert np.all(phases == 0x90)

        autd.send(Clear())

        for dev in autd.geometry:
            assert np.all(autd.link.modulation_buffer(dev.idx, Segment.S0) == 0xFF)
            intensities, phases = autd.link.drives_at(dev.idx, Segment.S0, 0)
            assert np.all(intensities == 0)
            assert np.all(phases == 0)


def test_synchronize():
    autd: Controller[Audit]
    with create_controller() as autd:
        autd.send(Synchronize())


def test_force_fan():
    autd: Controller[Audit]
    with create_controller() as autd:
        for dev in autd.geometry:
            assert not autd.link.is_force_fan(dev.idx)

        autd.send(ForceFan(lambda dev: dev.idx == 0))
        assert autd.link.is_force_fan(0)
        assert not autd.link.is_force_fan(1)

        autd.send(ForceFan(lambda dev: dev.idx == 1))
        assert not autd.link.is_force_fan(0)
        assert autd.link.is_force_fan(1)


def test_geometry():
    autd: Controller[Audit]
    with create_controller() as autd:
        assert autd.num_transducers == 249 * 2
        assert autd[0].num_transducers == 249


def test_tracing():
    tracing_init()
    autd: Controller[Audit]
    with create_controller() as autd:
        autd.close()



from pyautd3.gain.holo.constraint import EmissionConstraint

from .amplitude import Amplitude, Pa, dB
from .backend_nalgebra import NalgebraBackend
from .greedy import Greedy
from .gs import GS
from .gspat import GSPAT
from .lm import LM
from .naive import Naive

__all__ = [
    "dB",
    "Pa",
    "Amplitude",
    "NalgebraBackend",
    "EmissionConstraint",
    "GS",
    "GSPAT",
    "LM",
    "Greedy",
    "Naive",
]



from abc import ABCMeta, abstractmethod
from typing import Generic, Self, TypeVar

from pyautd3.derive import datagram
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, Segment, TransitionModeWrap

__all__ = []  # type: ignore[var-annotated]

DS = TypeVar("DS", bound="DatagramS")
P = TypeVar("P")


class DatagramS(Generic[P], metaclass=ABCMeta):
    @abstractmethod
    def _into_segment(self: Self, ptr: P, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramPtr:
        pass

    @abstractmethod
    def _raw_ptr(self: Self, geometry: Geometry) -> P:
        pass


@datagram
class DatagramWithSegment(Datagram, Generic[DS]):
    _datagram: DS
    _segment: Segment
    _transition_mode: TransitionModeWrap | None

    def __init__(self: Self, datagram: DS, segment: Segment, transition_mode: TransitionModeWrap | None) -> None:
        self._datagram = datagram
        self._segment = segment
        self._transition_mode = transition_mode

    def _datagram_ptr(self: Self, g: Geometry) -> DatagramPtr:
        raw_ptr = self._datagram._raw_ptr(g)
        return self._datagram._into_segment(raw_ptr, self._segment, self._transition_mode)



from collections.abc import Callable
from ctypes import CFUNCTYPE
from ctypes import c_uint8
from ctypes import c_uint16
from ctypes import c_void_p
from threading import Lock
from typing import Self
from pyautd3.derive import datagram
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.geometry import Geometry
from pyautd3.driver.geometry.device import Device
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr
from pyautd3.native_methods.autd3capi_driver import GeometryPtr
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold



class PulseWidthEncoder(Datagram):
    _cache: dict[int, Callable[[int], int]]
    _lock: Lock
    def __init__(self: PulseWidthEncoder, f: Callable[[Device], Callable[[int], int]] | None = None) -> None: ...
    def _datagram_ptr(self: PulseWidthEncoder, geometry: Geometry) -> DatagramPtr: ...
    def with_timeout(self: PulseWidthEncoder, timeout: timedelta | None) -> DatagramWithTimeout[PulseWidthEncoder]: ...
    def with_parallel_threshold(self: PulseWidthEncoder, threshold: int | None) -> DatagramWithParallelThreshold[PulseWidthEncoder]: ...



import ctypes
from collections.abc import Iterable
from typing import Self
import numpy as np
from pyautd3.derive import builder
from pyautd3.derive import datagram
from pyautd3.derive import gain
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.geometry import Geometry
from pyautd3.gain.holo.amplitude import Amplitude
from pyautd3.gain.holo.backend import Backend
from pyautd3.gain.holo.constraint import EmissionConstraint
from pyautd3.gain.holo.holo import HoloWithBackend
from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.native_methods.structs import Vector3
from pyautd3.gain.cache import Cache
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold
from pyautd3.native_methods.autd3capi_driver import Segment, TransitionModeWrap
from pyautd3.driver.datagram.with_segment import DatagramWithSegment



class LM(HoloWithBackend[LM]):
    def __init__(self: LM, backend: Backend, iterable: Iterable[tuple[np.ndarray, Amplitude]]) -> None: ...
    def _gain_ptr(self: LM, _: Geometry) -> GainPtr: ...
    def with_eps1(self: LM, eps1: float) -> LM: ...
    def with_eps2(self: LM, eps2: float) -> LM: ...
    def with_tau(self: LM, tau: float) -> LM: ...
    def with_kmax(self: LM, kmax: int) -> LM: ...
    def with_initial(self: LM, initial: np.ndarray) -> LM: ...
    def with_cache(self: LM, ) -> Cache[LM]: ...
    def with_timeout(self: LM, timeout: timedelta | None) -> DatagramWithTimeout[LM]: ...
    def with_parallel_threshold(self: LM, threshold: int | None) -> DatagramWithParallelThreshold[LM]: ...
    def with_segment(self: LM, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramWithSegment[LM]: ...
    @property
    def eps1(self: LM) -> float: ...
    @property
    def eps2(self: LM) -> float: ...
    @property
    def tau(self: LM) -> float: ...
    @property
    def kmax(self: LM) -> int: ...
    @property
    def initial(self: LM) -> np.ndarray: ...



import ctypes
from collections.abc import Iterable
from datetime import timedelta
from typing import Self
import numpy as np
from pyautd3.derive import datagram
from pyautd3.derive.derive_builder import builder
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.datagram.gain import Gain
from pyautd3.driver.datagram.stm.stm_sampling_config import STMSamplingConfig
from pyautd3.driver.datagram.with_segment import DatagramS
from pyautd3.driver.defined.freq import Freq
from pyautd3.driver.firmware.fpga import LoopBehavior
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.driver.firmware.fpga.transition_mode import TransitionMode
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr
from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.native_methods.autd3capi_driver import GainSTMMode
from pyautd3.native_methods.autd3capi_driver import GainSTMPtr
from pyautd3.native_methods.autd3capi_driver import Segment
from pyautd3.native_methods.autd3capi_driver import TransitionModeWrap
from pyautd3.native_methods.autd3capi_driver import LoopBehavior as _LoopBehavior
from pyautd3.native_methods.utils import _validate_ptr
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold
from pyautd3.native_methods.autd3capi_driver import Segment, TransitionModeWrap
from pyautd3.driver.datagram.with_segment import DatagramWithSegment



class GainSTM(DatagramS[GainSTMPtr], Datagram):
    _gains: np.ndarray
    _stm_sampling_config: STMSamplingConfig
    def __private_init__(self: GainSTM, sampling_config: STMSamplingConfig, gains: list[Gain]) -> None: ...
    def __init__(self: GainSTM, config: SamplingConfig | Freq[float] | timedelta, iterable: Iterable[Gain]) -> None: ...
    def _raw_ptr(self: GainSTM, geometry: Geometry) -> GainSTMPtr: ...
    def _ptr(self: GainSTM, gains: np.ndarray) -> GainSTMPtr: ...
    def _datagram_ptr(self: GainSTM, geometry: Geometry) -> DatagramPtr: ...
    def _into_segment(self: GainSTM, ptr: GainSTMPtr, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramPtr: ...
    def _sampling_config_intensity(self: GainSTM, ) -> SamplingConfig: ...
    def _sampling_config_phase(self: GainSTM, ) -> SamplingConfig: ...
    def with_mode(self: GainSTM, mode: GainSTMMode) -> GainSTM: ...
    def with_loop_behavior(self: GainSTM, loop_behavior: _LoopBehavior) -> GainSTM: ...
    def with_timeout(self: GainSTM, timeout: timedelta | None) -> DatagramWithTimeout[GainSTM]: ...
    def with_parallel_threshold(self: GainSTM, threshold: int | None) -> DatagramWithParallelThreshold[GainSTM]: ...
    def with_segment(self: GainSTM, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramWithSegment[GainSTM]: ...
    @classmethod
    def nearest(cls, config: Freq[float] | timedelta, iterable: Iterable[Gain]) -> GainSTM: ...
    @property
    def freq(self: GainSTM) -> Freq[float]: ...
    @property
    def period(self: GainSTM) -> timedelta: ...
    @property
    def sampling_config(self: GainSTM) -> SamplingConfig: ...
    @property
    def mode(self: GainSTM) -> GainSTMMode: ...
    @property
    def loop_behavior(self: GainSTM) -> _LoopBehavior: ...



from collections.abc import Iterable
from typing import Self
from pyautd3.derive import builder
from pyautd3.derive import datagram
from pyautd3.derive import modulation
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.modulation.sine import Sine
from pyautd3.native_methods.autd3capi_driver import ModulationPtr
from pyautd3.modulation.cache import Cache
from pyautd3.modulation.fir import Fir
from collections.abc import Iterable
from pyautd3.modulation.radiation_pressure import RadiationPressure
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold
from pyautd3.native_methods.autd3capi_driver import Segment, TransitionModeWrap
from pyautd3.driver.datagram.with_segment import DatagramWithSegment



class Fourier(Modulation):
    _components: list[Sine]
    def __init__(self: Fourier, iterable: Iterable[Sine]) -> None: ...
    def _modulation_ptr(self: Fourier, ) -> ModulationPtr: ...
    def with_clamp(self: Fourier, clamp: bool) -> Fourier: ...
    def with_scale_factor(self: Fourier, scale_factor: float | None) -> Fourier: ...
    def with_offset(self: Fourier, offset: int) -> Fourier: ...
    def with_cache(self: Fourier, ) -> Cache[Fourier]: ...
    def with_fir(self: Fourier, iterable: Iterable[float]) -> Fir[Fourier]: ...
    def with_radiation_pressure(self: Fourier, ) -> RadiationPressure[Fourier]: ...
    def with_timeout(self: Fourier, timeout: timedelta | None) -> DatagramWithTimeout[Fourier]: ...
    def with_parallel_threshold(self: Fourier, threshold: int | None) -> DatagramWithParallelThreshold[Fourier]: ...
    def with_segment(self: Fourier, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramWithSegment[Fourier]: ...
    @property
    def clamp(self: Fourier) -> bool: ...
    @property
    def scale_factor(self: Fourier) -> float | None: ...
    @property
    def offset(self: Fourier) -> int: ...



import ctypes
from collections.abc import Iterable
from typing import Generic, Self, TypeVar

import numpy as np

from pyautd3.derive import datagram, modulation
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import ModulationPtr

M = TypeVar("M", bound=Modulation)


@datagram
@datagram_with_segment
@modulation
class Fir(Modulation, Generic[M]):
    _m: M
    _coef: np.ndarray

    def __init__(self: Self, m: M, iterable: Iterable[float]) -> None:
        self._m = m
        self._loop_behavior = m._loop_behavior
        self._coef = np.fromiter(iterable, dtype=ctypes.c_float)

    def _modulation_ptr(self: Self) -> ModulationPtr:
        return Base().modulation_with_fir(
            self._m._modulation_ptr(),
            self._loop_behavior,
            self._coef.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),  # type: ignore[arg-type]
            len(self._coef),
        )



import ctypes
from datetime import timedelta
from typing import Self

import numpy as np

from pyautd3.driver.link import Link, LinkBuilder
from pyautd3.native_methods.autd3_driver import Drive as Drive_
from pyautd3.native_methods.autd3capi import NativeMethods as LinkAudit
from pyautd3.native_methods.autd3capi_driver import ControllerPtr, HandlePtr, LinkBuilderPtr, LinkPtr, LoopBehavior, Segment, SilencerTarget

__all__ = []  # type: ignore[var-annotated]


class Audit(Link):
    _ptr: LinkPtr

    class _Builder(LinkBuilder["Audit"]):
        def _link_builder_ptr(self: Self) -> LinkBuilderPtr:
            return LinkAudit().link_audit()

        def _resolve_link(self: Self, handle: HandlePtr, ptr: ControllerPtr) -> "Audit":
            return Audit(handle, LinkAudit().link_get(ptr))

    def __init__(self: Self, handle: HandlePtr, ptr: LinkPtr) -> None:
        super().__init__(handle, ptr)

    @staticmethod
    def builder() -> _Builder:
        return Audit._Builder()

    def down(self: Self) -> None:
        LinkAudit().link_audit_down(self._ptr)

    def up(self: Self) -> None:
        LinkAudit().link_audit_up(self._ptr)

    def is_open(self: Self) -> bool:
        return bool(LinkAudit().link_audit_is_open(self._ptr))

    def is_force_fan(self: Self, idx: int) -> bool:
        return bool(LinkAudit().link_audit_fpga_is_force_fan(self._ptr, idx))

    def break_down(self: Self) -> None:
        LinkAudit().link_audit_break_down(self._ptr)

    def repair(self: Self) -> None:
        LinkAudit().link_audit_repair(self._ptr)

    def last_timeout(self: Self) -> timedelta | None:
        ns = int(LinkAudit().link_audit_last_timeout(self._ptr))
        return timedelta(microseconds=ns / 1000) if ns >= 0 else None

    def last_parallel_threshold(self: Self) -> int | None:
        threshold = int(LinkAudit().link_audit_last_parallel_threshold(self._ptr))
        return threshold if threshold >= 0 else None

    def silencer_strict_mode(self: Self, idx: int) -> bool:
        return bool(LinkAudit().link_audit_cpu_silencer_strict_mode(self._ptr, idx))

    def silencer_update_rate_intensity(self: Self, idx: int) -> int:
        return int(LinkAudit().link_audit_fpga_silencer_update_rate_intensity(self._ptr, idx))

    def silencer_update_rate_phase(self: Self, idx: int) -> int:
        return int(LinkAudit().link_audit_fpga_silencer_update_rate_phase(self._ptr, idx))

    def silencer_completion_steps_intensity(self: Self, idx: int) -> int:
        return int(LinkAudit().link_audit_fpga_silencer_completion_steps_intensity(self._ptr, idx))

    def silencer_completion_steps_phase(self: Self, idx: int) -> int:
        return int(LinkAudit().link_audit_fpga_silencer_completion_steps_phase(self._ptr, idx))

    def silencer_fixed_completion_steps_mode(self: Self, idx: int) -> bool:
        return bool(LinkAudit().link_audit_fpga_silencer_fixed_completion_steps_mode(self._ptr, idx))

    def silencer_target(self: Self, idx: int) -> SilencerTarget:
        return LinkAudit().link_audit_fpga_silencer_target(self._ptr, idx)

    def debug_types(self: Self, idx: int) -> np.ndarray:
        buf = np.zeros([4]).astype(ctypes.c_uint8)
        LinkAudit().link_audit_fpga_debug_types(self._ptr, idx, np.ctypeslib.as_ctypes(buf))
        return buf

    def debug_values(self: Self, idx: int) -> np.ndarray:
        buf = np.zeros([4]).astype(ctypes.c_uint64)
        LinkAudit().link_audit_fpga_debug_values(self._ptr, idx, np.ctypeslib.as_ctypes(buf))
        return buf

    def assert_thermal_sensor(self: Self, idx: int) -> None:
        LinkAudit().link_audit_fpga_assert_thermal_sensor(self._ptr, idx)

    def deassert_thermal_sensor(self: Self, idx: int) -> None:
        LinkAudit().link_audit_fpga_deassert_thermal_sensor(self._ptr, idx)

    def modulation_buffer(
        self: Self,
        idx: int,
        segment: Segment,
    ) -> np.ndarray:
        n = int(LinkAudit().link_audit_fpga_modulation_cycle(self._ptr, segment, idx))
        buf = np.zeros([n]).astype(ctypes.c_uint8)
        LinkAudit().link_audit_fpga_modulation_buffer(self._ptr, segment, idx, np.ctypeslib.as_ctypes(buf), n)
        return buf

    def modulation_frequency_division(self: Self, idx: int, segment: Segment) -> int:
        return int(LinkAudit().link_audit_fpga_modulation_freq_division(self._ptr, segment, idx))

    def modulation_loop_behavior(self: Self, idx: int, segment: Segment) -> LoopBehavior:
        return LinkAudit().link_audit_fpga_modulation_loop_behavior(self._ptr, segment, idx)

    def drives_at(self: Self, idx: int, segment: Segment, stm_idx: int) -> tuple[np.ndarray, np.ndarray]:
        n = int(LinkAudit().link_audit_cpu_num_transducers(self._ptr, idx))
        drive = np.zeros(n, dtype=Drive_)
        LinkAudit().link_audit_fpga_drives_at(
            self._ptr,
            segment,
            idx,
            stm_idx,
            drive.ctypes.data_as(ctypes.POINTER(Drive_)),  # type: ignore[arg-type]
        )
        return np.array([int(d[1]) for d in drive]), np.array([int(d[0]) for d in drive])

    def sound_speed(self: Self, idx: int, segment: Segment) -> int:
        return int(LinkAudit().link_audit_fpga_sound_speed(self._ptr, segment, idx))

    def stm_cycle(self: Self, idx: int, segment: Segment) -> int:
        return int(LinkAudit().link_audit_fpga_stm_cycle(self._ptr, segment, idx))

    def is_stm_gain_mode(self: Self, idx: int, segment: Segment) -> int:
        return int(LinkAudit().link_audit_fpga_is_stm_gain_mode(self._ptr, segment, idx))

    def stm_freqency_division(self: Self, idx: int, segment: Segment) -> int:
        return int(LinkAudit().link_audit_fpga_stm_freq_division(self._ptr, segment, idx))

    def stm_loop_behavior(self: Self, idx: int, segment: Segment) -> LoopBehavior:
        return LinkAudit().link_audit_fpga_stm_loop_behavior(self._ptr, segment, idx)

    def current_stm_segment(self: Self, idx: int) -> Segment:
        return LinkAudit().link_audit_fpga_current_stm_segment(self._ptr, idx)

    def current_mod_segment(self: Self, idx: int) -> Segment:
        return LinkAudit().link_audit_fpga_current_mod_segment(self._ptr, idx)

    def pulse_width_encoder_table(self: Self, idx: int) -> np.ndarray:
        p = np.zeros([256]).astype(ctypes.c_uint8)
        LinkAudit().link_audit_fpga_pulse_width_encoder_table(
            self._ptr,
            idx,
            np.ctypeslib.as_ctypes(p),
        )
        return p



import ctypes
from collections.abc import Iterable
from typing import Generic
from typing import Self
from typing import TypeVar
import numpy as np
from pyautd3.derive import datagram
from pyautd3.derive import modulation
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import ModulationPtr
from pyautd3.modulation.cache import Cache
from pyautd3.modulation.radiation_pressure import RadiationPressure
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold
from pyautd3.native_methods.autd3capi_driver import Segment, TransitionModeWrap
from pyautd3.driver.datagram.with_segment import DatagramWithSegment

M = TypeVar("M", bound=Modulation)

class Fir(Modulation, Generic[M]):
    _m: M
    _coef: np.ndarray
    def __init__(self: Fir[M], m: M, iterable: Iterable[float]) -> None: ...
    def _modulation_ptr(self: Fir[M], ) -> ModulationPtr: ...
    def with_cache(self: Fir[M], ) -> Cache[Fir[M]]: ...
    def with_radiation_pressure(self: Fir[M], ) -> RadiationPressure[Fir[M]]: ...
    def with_timeout(self: Fir[M], timeout: timedelta | None) -> DatagramWithTimeout[Fir[M]]: ...
    def with_parallel_threshold(self: Fir[M], threshold: int | None) -> DatagramWithParallelThreshold[Fir[M]]: ...
    def with_segment(self: Fir[M], segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramWithSegment[Fir[M]]: ...



from typing import Self

import numpy as np
from numpy.typing import ArrayLike

import pyautd3.native_methods.autd3capi_driver as consts
from pyautd3.derive import builder


@builder
class AUTD3:
    _prop_position: np.ndarray
    _param_rotation: np.ndarray

    def __init__(self: Self, pos: ArrayLike) -> None:
        self._prop_position = np.array(pos)
        self._param_rotation = np.array([1, 0, 0, 0])

    TRANS_SPACING: float = consts.TRANS_SPACING_MM
    DEVICE_WIDTH: float = consts.DEVICE_WIDTH_MM
    DEVICE_HEIGHT: float = consts.DEVICE_HEIGHT_MM
    NUM_TRANS_IN_X: int = consts.NUM_TRANS_IN_X
    NUM_TRANS_IN_Y: int = consts.NUM_TRANS_IN_Y
    NUM_TRANS_IN_UNIT: int = consts.NUM_TRANS_IN_UNIT



from abc import ABCMeta, abstractmethod
from datetime import timedelta
from typing import Self, TypeVar

from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.datagram.with_segment import DatagramS
from pyautd3.driver.defined.freq import Freq
from pyautd3.driver.firmware.fpga import LoopBehavior
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.driver.firmware.fpga.transition_mode import TransitionMode
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, ModulationPtr, Segment, TransitionModeWrap
from pyautd3.native_methods.autd3capi_driver import LoopBehavior as _LoopBehavior

__all__ = []  # type: ignore[var-annotated]

M = TypeVar("M", bound="Modulation")


class Modulation(
    DatagramS[ModulationPtr],
    Datagram,
    metaclass=ABCMeta,
):
    _loop_behavior: _LoopBehavior

    def __init__(self: Self) -> None:
        super().__init__()
        self._loop_behavior = LoopBehavior.Infinite

    def _raw_ptr(self: Self, _: Geometry) -> ModulationPtr:
        return self._modulation_ptr()

    def _datagram_ptr(self: Self, _: Geometry) -> DatagramPtr:
        return Base().modulation_into_datagram(self._modulation_ptr())

    def _into_segment(
        self: Self,
        ptr: ModulationPtr,
        segment: Segment,
        transition_mode: TransitionModeWrap | None,
    ) -> DatagramPtr:
        return Base().modulation_into_datagram_with_segment(
            ptr,
            segment,
            transition_mode if transition_mode is not None else TransitionMode.NONE,
        )

    @abstractmethod
    def _modulation_ptr(self: Self) -> ModulationPtr:
        pass

    def with_loop_behavior(self: Self, loop_behavior: _LoopBehavior) -> Self:
        self._loop_behavior = loop_behavior
        return self

    @property
    def loop_behavior(self: Self) -> _LoopBehavior:
        return self._loop_behavior

    @property
    def sampling_config(self: Self) -> SamplingConfig:
        return SamplingConfig(Base().modulation_sampling_config(self._modulation_ptr()))

    def _sampling_config_intensity(self: Self) -> SamplingConfig:
        return self.sampling_config

    def _sampling_config_phase(self: Self) -> SamplingConfig:
        return SamplingConfig(0xFFFF)


class ModulationWithSamplingConfig(Modulation, metaclass=ABCMeta):
    _config: SamplingConfig

    def __init__(self: Self, config: SamplingConfig | Freq[int] | Freq[float] | timedelta) -> None:
        super().__init__()
        self._config = SamplingConfig(config)

    def with_sampling_config(self: Self, config: SamplingConfig | Freq[int] | Freq[float] | timedelta) -> Self:
        self._config = SamplingConfig(config)
        return self



from datetime import timedelta

from pyautd3.native_methods.autd3_driver import Segment
from pyautd3.native_methods.autd3capi_driver import TransitionModeWrap


def datagram(cls):
    def with_timeout(self, timeout: timedelta | None):
        from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout

        return DatagramWithTimeout(self, timeout)

    def with_parallel_threshold(self, threshold: int | None):
        from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold

        return DatagramWithParallelThreshold(self, threshold)

    cls.with_timeout = with_timeout  # type: ignore[attr-defined]
    cls.with_parallel_threshold = with_parallel_threshold  # type: ignore[attr-defined]

    return cls


def datagram_with_segment(cls):
    def with_segment(self, segment: Segment, transition_mode: TransitionModeWrap | None):
        from pyautd3.driver.datagram.with_segment import DatagramWithSegment

        return DatagramWithSegment(self, segment, transition_mode)

    cls.with_segment = with_segment  # type: ignore[attr-defined]

    return cls



from typing import Self
from pyautd3.derive import datagram
from pyautd3.derive import gain
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.gain import Gain
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.gain.cache import Cache
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold
from pyautd3.native_methods.autd3capi_driver import Segment, TransitionModeWrap
from pyautd3.driver.datagram.with_segment import DatagramWithSegment



class Null(Gain):
    def __init__(self: Null, ) -> None: ...
    def _gain_ptr(self: Null, _: Geometry) -> GainPtr: ...
    def with_cache(self: Null, ) -> Cache[Null]: ...
    def with_timeout(self: Null, timeout: timedelta | None) -> DatagramWithTimeout[Null]: ...
    def with_parallel_threshold(self: Null, threshold: int | None) -> DatagramWithParallelThreshold[Null]: ...
    def with_segment(self: Null, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramWithSegment[Null]: ...



import ctypes
from collections.abc import Callable
from typing import Self

from pyautd3.derive import datagram
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.geometry import Device, Geometry, Transducer
from pyautd3.ethercat.dc_sys_time import DcSysTime
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, DebugTypeWrap, GeometryPtr, GPIOOut
from pyautd3.native_methods.utils import ConstantADT

__all__ = ["DebugType"]


class DebugType(metaclass=ConstantADT):
    def __new__(cls: type["DebugType"]) -> "DebugType":
        raise NotImplementedError

    NONE: DebugTypeWrap = Base().debug_type_none()
    BaseSignal: DebugTypeWrap = Base().debug_type_base_signal()
    Thermo: DebugTypeWrap = Base().debug_type_thermo()
    ForceFan: DebugTypeWrap = Base().debug_type_force_fan()
    Sync: DebugTypeWrap = Base().debug_type_sync()
    ModSegment: DebugTypeWrap = Base().debug_type_mod_segment()

    @staticmethod
    def ModIdx(idx: int) -> DebugTypeWrap:  # noqa: N802
        return Base().debug_type_mod_idx(idx)

    StmSegment: DebugTypeWrap = Base().debug_type_stm_segment()

    @staticmethod
    def StmIdx(idx: int) -> DebugTypeWrap:  # noqa: N802
        return Base().debug_type_stm_idx(idx)

    IsStmMode: DebugTypeWrap = Base().debug_type_is_stm_mode()

    @staticmethod
    def PwmOut(tr: Transducer) -> DebugTypeWrap:  # noqa: N802
        return Base().debug_type_pwm_out(tr._ptr)

    @staticmethod
    def Direct(value: bool) -> DebugTypeWrap:  # noqa: N802, FBT001
        return Base().debug_type_direct(value)

    @staticmethod
    def SysTimeEq(value: DcSysTime) -> DebugTypeWrap:  # noqa: N802
        return Base().debug_type_sys_time_eq(value.sys_time)


@datagram
class DebugSettings(Datagram):
    def __init__(self: Self, f: Callable[[Device, GPIOOut], DebugTypeWrap]) -> None:
        super().__init__()

        def f_native(_context: ctypes.c_void_p, geometry_ptr: GeometryPtr, dev_idx: int, gpio: GPIOOut, res) -> None:  # noqa: ANN001
            res[0] = f(Device(dev_idx, geometry_ptr), gpio)

        self._f_native = ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            GeometryPtr,
            ctypes.c_uint32,
            ctypes.c_uint8,
            ctypes.POINTER(DebugTypeWrap),
        )(f_native)

    def _datagram_ptr(self: Self, geometry: Geometry) -> DatagramPtr:
        return Base().datagram_debug_settings(self._f_native, None, geometry._geometry_ptr)  # type: ignore[arg-type]



import platform
from typing import Self

from pyautd3.derive import builder
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import SpinStrategyTag as SpinStrategy
from pyautd3.native_methods.autd3capi_driver import TimerStrategyWrap
from pyautd3.native_methods.utils import ConstantADT


@builder
class StdSleeper:
    _prop_timer_resolution: int | None

    def __init__(self: Self, *, timer_resolution: int | None = 1) -> None:
        self._prop_timer_resolution = timer_resolution


@builder
class SpinSleeper:
    _prop_native_accuracy_ns: int
    _param_spin_strategy: SpinStrategy

    def __init__(
        self: Self,
        native_accuracy_ns: int | None = None,
    ) -> None:
        self._prop_native_accuracy_ns = native_accuracy_ns if native_accuracy_ns is not None else int(Base().timer_strategy_spin_default_accuracy())
        self._param_spin_strategy = SpinStrategy.SpinLoopHint if platform.system() == "Windows" else SpinStrategy.YieldThread


@builder
class AsyncSleeper:
    _prop_timer_resolution: int | None

    def __init__(self: Self, *, timer_resolution: int | None = 1) -> None:
        self._prop_timer_resolution = timer_resolution


class WaitableSleeper:
    def __init__(self: Self) -> None:
        pass


class TimerStrategy(metaclass=ConstantADT):
    @staticmethod
    def Std(sleeper: StdSleeper) -> TimerStrategyWrap:  # noqa: N802
        return Base().timer_strategy_std(sleeper._prop_timer_resolution if sleeper._prop_timer_resolution is not None else 0)

    @staticmethod
    def Spin(sleeper: SpinSleeper) -> TimerStrategyWrap:  # noqa: N802
        return Base().timer_strategy_spin(sleeper._prop_native_accuracy_ns, sleeper._param_spin_strategy)

    @staticmethod
    def Async(sleeper: AsyncSleeper) -> TimerStrategyWrap:  # noqa: N802
        return Base().timer_strategy_async(sleeper.timer_resolution if sleeper.timer_resolution is not None else 0)

    @staticmethod
    def Waitable(_sleeper: WaitableSleeper) -> TimerStrategyWrap:  # noqa: N802
        return Base().timer_strategy_waitable()

    def __new__(cls: type["TimerStrategy"]) -> "TimerStrategy":
        raise NotImplementedError



from typing import Self
from typing import TypeVar
from pyautd3.derive import builder
from pyautd3.derive import datagram
from pyautd3.derive import modulation
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.modulation import ModulationWithSamplingConfig
from pyautd3.driver.defined.angle import Angle
from pyautd3.driver.defined.angle import rad
from pyautd3.driver.defined.freq import Freq
from pyautd3.driver.defined.freq import Hz
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.modulation.sampling_mode import ISamplingMode
from pyautd3.modulation.sampling_mode import SamplingModeExact
from pyautd3.modulation.sampling_mode import SamplingModeExactFloat
from pyautd3.modulation.sampling_mode import SamplingModeNearest
from pyautd3.native_methods.autd3capi_driver import ModulationPtr
from pyautd3.modulation.cache import Cache
from pyautd3.modulation.fir import Fir
from collections.abc import Iterable
from pyautd3.modulation.radiation_pressure import RadiationPressure
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold
from pyautd3.native_methods.autd3capi_driver import Segment, TransitionModeWrap
from pyautd3.driver.datagram.with_segment import DatagramWithSegment

T = TypeVar("T", int, float)

class Sine(ModulationWithSamplingConfig):
    _mode: ISamplingMode
    def __private__init__(self: Sine, mode: ISamplingMode) -> None: ...
    def __init__(self: Sine, freq: Freq[T]) -> None: ...
    def _modulation_ptr(self: Sine, ) -> ModulationPtr: ...
    def with_intensity(self: Sine, intensity: int) -> Sine: ...
    def with_offset(self: Sine, offset: int) -> Sine: ...
    def with_phase(self: Sine, phase: Angle) -> Sine: ...
    def with_clamp(self: Sine, clamp: bool) -> Sine: ...
    def with_cache(self: Sine, ) -> Cache[Sine]: ...
    def with_fir(self: Sine, iterable: Iterable[float]) -> Fir[Sine]: ...
    def with_radiation_pressure(self: Sine, ) -> RadiationPressure[Sine]: ...
    def with_timeout(self: Sine, timeout: timedelta | None) -> DatagramWithTimeout[Sine]: ...
    def with_parallel_threshold(self: Sine, threshold: int | None) -> DatagramWithParallelThreshold[Sine]: ...
    def with_segment(self: Sine, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramWithSegment[Sine]: ...
    @classmethod
    def nearest(cls, freq: Freq[float]) -> Sine: ...
    @property
    def freq(self: Sine) -> Freq[int] | Freq[float]: ...
    @property
    def intensity(self: Sine) -> int: ...
    @property
    def offset(self: Sine) -> int: ...
    @property
    def phase(self: Sine) -> Angle: ...
    @property
    def clamp(self: Sine) -> bool: ...



from datetime import timedelta
from typing import Generic
from typing import Self
from typing import TypeVar
from pyautd3.derive import datagram
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold

D = TypeVar("D", bound="Datagram")

class DatagramWithTimeout(Datagram, Generic[D]):
    _datagram: D
    _timeout: timedelta | None
    def __init__(self: DatagramWithTimeout[D], datagram: D, timeout: timedelta | None) -> None: ...
    def _datagram_ptr(self: DatagramWithTimeout[D], g: Geometry) -> DatagramPtr: ...
    def with_parallel_threshold(self: DatagramWithTimeout[D], threshold: int | None) -> DatagramWithParallelThreshold[DatagramWithTimeout[D]]: ...



from collections.abc import Iterable


def modulation(cls):
    def with_cache(self):
        from pyautd3.modulation.cache import Cache

        return Cache(self)

    def with_fir(self, iterable: Iterable[float]):
        from pyautd3.modulation.fir import Fir

        return Fir(self, iterable)

    def with_radiation_pressure(self):
        from pyautd3.modulation.radiation_pressure import RadiationPressure

        return RadiationPressure(self)

    cls.with_cache = with_cache  # type: ignore[attr-defined]
    cls.with_fir = with_fir  # type: ignore[attr-defined]
    cls.with_radiation_pressure = with_radiation_pressure  # type: ignore[attr-defined]

    return cls



from typing import Self

import numpy as np
from numpy.typing import ArrayLike

from pyautd3.derive import builder, datagram, gain
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.gain import Gain
from pyautd3.driver.firmware.fpga.emit_intensity import EmitIntensity
from pyautd3.driver.firmware.fpga.phase import Phase
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.native_methods.structs import Vector3


@gain
@datagram
@datagram_with_segment
@builder
class Plane(Gain):
    _prop_dir: np.ndarray
    _param_intensity: EmitIntensity
    _param_phase_offset: Phase

    def __init__(self: Self, direction: ArrayLike) -> None:
        super().__init__()
        self._prop_dir = np.array(direction)
        self._param_intensity = EmitIntensity.maximum()
        self._param_phase_offset = Phase(0)

    def _gain_ptr(self: Self, _: Geometry) -> GainPtr:
        return Base().gain_plane(Vector3(self._prop_dir), self._param_intensity.value, self._param_phase_offset.value)
